# -*- test-case-name: wokkel.test.test_generic -*-
#
# Copyright (c) Ralph Meijer.
# See LICENSE for details.

"""
Generic XMPP protocol helpers.
"""

from __future__ import division, absolute_import

from zope.interface import implementer

from twisted.internet import defer, protocol
from twisted.python import reflect
from twisted.words.protocols.jabber import error, jid, xmlstream
from twisted.words.protocols.jabber.xmlstream import toResponse
from twisted.words.xish import domish, utility
from twisted.words.xish.xmlstream import BootstrapMixin

from wokkel.iwokkel import IDisco
from wokkel.subprotocols import XMPPHandler

IQ_GET = '/iq[@type="get"]'
IQ_SET = '/iq[@type="set"]'

NS_VERSION = "jabber:iq:version"
VERSION = IQ_GET + '/query[@xmlns="' + NS_VERSION + '"]'


def parseXml(string):
    """
    Parse serialized XML into a DOM structure.

    @param string: The serialized XML to be parsed, UTF-8 encoded.
    @type string: L{str}.
    @return: The DOM structure, or C{None} on empty or incomplete input.
    @rtype: L{domish.Element}
    """
    roots = []
    results = []
    elementStream = domish.elementStream()
    elementStream.DocumentStartEvent = roots.append
    elementStream.ElementEvent = lambda elem: roots[0].addChild(elem)
    elementStream.DocumentEndEvent = lambda: results.append(roots[0])
    elementStream.parse(string)
    return results and results[0] or None


def stripNamespace(rootElement):
    namespace = rootElement.uri

    def strip(element):
        if element.uri == namespace:
            element.uri = None
            if element.defaultUri == namespace:
                element.defaultUri = None
            for child in element.elements():
                strip(child)

    if namespace is not None:
        strip(rootElement)

    return rootElement


class FallbackHandler(XMPPHandler):
    """
    XMPP subprotocol handler that catches unhandled iq requests.

    Unhandled iq requests are replied to with a service-unavailable stanza
    error.
    """

    def connectionInitialized(self):
        self.xmlstream.addObserver(IQ_SET, self.iqFallback, -1)
        self.xmlstream.addObserver(IQ_GET, self.iqFallback, -1)

    def iqFallback(self, iq):
        if iq.handled == True:
            return

        reply = error.StanzaError("service-unavailable")
        self.xmlstream.send(reply.toResponse(iq))


@implementer(IDisco)
class VersionHandler(XMPPHandler):
    """
    XMPP subprotocol handler for XMPP Software Version.

    This protocol is described in
    U{XEP-0092<http://xmpp.org/extensions/xep-0092.html>}.
    """

    def __init__(self, name, version):
        self.name = name
        self.version = version

    def connectionInitialized(self):
        self.xmlstream.addObserver(VERSION, self.onVersion)

    def onVersion(self, iq):
        response = toResponse(iq, "result")

        query = response.addElement((NS_VERSION, "query"))
        query.addElement("name", content=self.name)
        query.addElement("version", content=self.version)
        self.send(response)

        iq.handled = True

    def getDiscoInfo(self, requestor, target, nodeIdentifier=""):
        info = set()

        if not nodeIdentifier:
            from wokkel import disco

            info.add(disco.DiscoFeature(NS_VERSION))

        return defer.succeed(info)

    def getDiscoItems(self, requestor, target, nodeIdentifier=""):
        return defer.succeed([])


class XmlPipe(object):
    """
    XML stream pipe.

    Connects two objects that communicate stanzas through an XML stream like
    interface. Each of the ends of the pipe (sink and source) can be used to
    send XML stanzas to the other side, or add observers to process XML stanzas
    that were sent from the other side.

    XML pipes are usually used in place of regular XML streams that are
    transported over TCP. This is the reason for the use of the names source
    and sink for both ends of the pipe. The source side corresponds with the
    entity that initiated the TCP connection, whereas the sink corresponds with
    the entity that accepts that connection. In this object, though, the source
    and sink are treated equally.

    Unlike Jabber
    L{XmlStream<twisted.words.protocols.jabber.xmlstream.XmlStream>}s, the sink
    and source objects are assumed to represent an eternal connected and
    initialized XML stream. As such, events corresponding to connection,
    disconnection, initialization and stream errors are not dispatched or
    processed.

    @ivar source: Source XML stream.
    @ivar sink: Sink XML stream.
    """

    def __init__(self):
        self.source = utility.EventDispatcher()
        self.sink = utility.EventDispatcher()
        self.source.send = lambda obj: self.sink.dispatch(obj)
        self.sink.send = lambda obj: self.source.dispatch(obj)


class Stanza(object):
    """
    Abstract representation of a stanza.

    @ivar sender: The sending entity.
    @type sender: L{jid.JID}
    @ivar recipient: The receiving entity.
    @type recipient: L{jid.JID}
    """

    recipient = None
    sender = None
    stanzaKind = None
    stanzaID = None
    stanzaType = None

    def __init__(self, recipient=None, sender=None):
        self.recipient = recipient
        self.sender = sender

    @classmethod
    def fromElement(Class, element):
        """
        Create a stanza from a L{domish.Element}.
        """
        stanza = Class()
        stanza.parseElement(element)
        return stanza

    def parseElement(self, element):
        """
        Parse the stanza element.

        This is called with the stanza's element when a L{Stanza} is
        created using L{fromElement}. It parses the stanza's core attributes
        (addressing, type and id), strips the namespace from the stanza
        element for easier transport across streams and passes on
        child elements for further parsing.

        Child element parsers are defined by providing a C{childParsers}
        attribute on a subclass, as a mapping from (URI, name) to the name
        of the handler on C{self}. C{parseElement} will accumulate
        C{childParsers} from its class hierarchy, iterate over the child
        elements and pass it to matching handlers based on the child element's
        URI and name. The special key of C{None} can be used to pass all
        child elements to.
        """
        if element.hasAttribute("from"):
            self.sender = jid.internJID(element["from"])
        if element.hasAttribute("to"):
            self.recipient = jid.internJID(element["to"])
        self.stanzaType = element.getAttribute("type")
        self.stanzaID = element.getAttribute("id")

        # Save element
        stripNamespace(element)
        self.element = element

        # accumulate all childHandlers in the class hierarchy of Class
        handlers = {}
        reflect.accumulateClassDict(self.__class__, "childParsers", handlers)

        for child in element.elements():
            try:
                handler = handlers[child.uri, child.name]
            except KeyError:
                try:
                    handler = handlers[None]
                except KeyError:
                    continue

            getattr(self, handler)(child)

    def toElement(self):
        element = domish.Element((None, self.stanzaKind))
        if self.sender is not None:
            element["from"] = self.sender.full()
        if self.recipient is not None:
            element["to"] = self.recipient.full()
        if self.stanzaType:
            element["type"] = self.stanzaType
        if self.stanzaID:
            element["id"] = self.stanzaID
        return element


class ErrorStanza(Stanza):

    def parseElement(self, element):
        Stanza.parseElement(self, element)
        self.exception = error.exceptionFromStanza(element)


class Request(Stanza):
    """
    IQ request stanza.

    This is a base class for IQ get or set stanzas, to be used with
    L{wokkel.subprotocols.StreamManager.request}.
    """

    stanzaKind = "iq"
    stanzaType = "get"
    timeout = None

    childParsers = {None: "parseRequest"}

    def __init__(self, recipient=None, sender=None, stanzaType=None):
        Stanza.__init__(self, recipient=recipient, sender=sender)
        if stanzaType is not None:
            self.stanzaType = stanzaType

    def parseRequest(self, element):
        """
        Called with the request's child element for parsing.

        When a request instance is created using L{fromElement}, this method
        is called with the child element of the iq. Override this method for
        parsing the request's payload.
        """

    def toElement(self):
        element = Stanza.toElement(self)

        if not self.stanzaID:
            element.addUniqueId()
            self.stanzaID = element["id"]

        return element


class DeferredXmlStreamFactory(BootstrapMixin, protocol.ClientFactory):
    protocol = xmlstream.XmlStream

    def __init__(self, authenticator):
        BootstrapMixin.__init__(self)

        self.authenticator = authenticator

        deferred = defer.Deferred()
        self.deferred = deferred
        self.addBootstrap(xmlstream.STREAM_AUTHD_EVENT, self.deferred.callback)
        self.addBootstrap(xmlstream.INIT_FAILED_EVENT, deferred.errback)

    def buildProtocol(self, addr):
        """
        Create an instance of XmlStream.

        A new authenticator instance will be created and passed to the new
        XmlStream. Registered bootstrap event observers are installed as well.
        """
        xs = self.protocol(self.authenticator)
        xs.factory = self
        self.installBootstraps(xs)
        return xs

    def clientConnectionFailed(self, connector, reason):
        self.deferred.errback(reason)



# -*- test-case-name: wokkel.test.test_shim -*-
#
# Copyright (c) Ralph Meijer.
# See LICENSE for details.

"""
Tests for {wokkel.shim}.
"""

from __future__ import division, absolute_import

from twisted.trial import unittest

from wokkel import shim
from wokkel.generic import parseXml

NS_SHIM = "http://jabber.org/protocol/shim"


class HeadersTest(unittest.TestCase):
    """
    Tests for L{wokkel.shim.headers}.
    """

    def test_noHeaders(self):
        headers = shim.Headers([])
        self.assertEquals(NS_SHIM, headers.uri)
        self.assertEquals("headers", headers.name)
        self.assertEquals([], headers.children)

    def test_header(self):
        headers = shim.Headers([("Urgency", "high")])
        elements = list(headers.elements())
        self.assertEquals(1, len(elements))
        header = elements[0]
        self.assertEquals(NS_SHIM, header.uri)
        self.assertEquals("header", header.name)
        self.assertEquals("Urgency", header["name"])
        self.assertEquals("high", str(header))

    def test_headerRepeated(self):
        """
        Some headers can appear more than once with the same name.
        """
        headers = shim.Headers([("Collection", "node1"), ("Collection", "node2")])
        elements = list(headers.elements())
        self.assertEquals(2, len(elements))
        collections = set(
            (str(element) for element in elements if element["name"] == "Collection")
        )
        self.assertIn("node1", collections)
        self.assertIn("node2", collections)


class ExtractHeadersTest(unittest.TestCase):
    """
    Tests for L{wokkel.shim.extractHeaders}.
    """

    def test_noHeaders(self):
        """
        A stanza without headers results in an empty dictionary.
        """
        stanza = parseXml("""<message/>""")
        headers = shim.extractHeaders(stanza)
        self.assertEquals({}, headers)

    def test_headers(self):
        """
        A stanza with headers results in a dictionary with those headers.
        """
        xml = """<message>
                   <headers xmlns='http://jabber.org/protocol/shim'>
                     <header name='Collection'>node1</header>
                     <header name='Urgency'>high</header>
                   </headers>
                 </message>"""

        stanza = parseXml(xml)
        headers = shim.extractHeaders(stanza)
        self.assertEquals({"Urgency": ["high"], "Collection": ["node1"]}, headers)

    def test_headersRepeated(self):
        """
        Some headers may appear repeatedly. Make sure all values are extracted.
        """
        xml = """<message>
                   <headers xmlns='http://jabber.org/protocol/shim'>
                     <header name='Collection'>node1</header>
                     <header name='Urgency'>high</header>
                     <header name='Collection'>node2</header>
                   </headers>
                 </message>"""

        stanza = parseXml(xml)
        headers = shim.extractHeaders(stanza)
        self.assertEquals(
            {"Urgency": ["high"], "Collection": ["node1", "node2"]}, headers
        )



# -*- test-case-name: wokkel.test.test_subprotocols -*-
#
# Copyright (c) Twisted Matrix Laboratories.
# See LICENSE for details.

"""
XMPP subprotocol support.
"""

from __future__ import division, absolute_import

from zope.interface import implementer

from twisted.internet import defer
from twisted.internet.error import ConnectionDone
from twisted.python import failure, log
from twisted.words.protocols.jabber import error, ijabber, xmlstream
from twisted.words.protocols.jabber.xmlstream import toResponse
from twisted.words.protocols.jabber.xmlstream import XMPPHandlerCollection
from twisted.words.xish import xpath
from twisted.words.xish.domish import IElement


@implementer(ijabber.IXMPPHandler)
class XMPPHandler(object):
    """
    XMPP protocol handler.

    Classes derived from this class implement (part of) one or more XMPP
    extension protocols, and are referred to as a subprotocol implementation.
    """

    def __init__(self):
        self.parent = None
        self.xmlstream = None

    def setHandlerParent(self, parent):
        self.parent = parent
        self.parent.addHandler(self)

    def disownHandlerParent(self, parent):
        self.parent.removeHandler(self)
        self.parent = None

    def makeConnection(self, xs):
        self.xmlstream = xs
        self.connectionMade()

    def connectionMade(self):
        """
        Called after a connection has been established.

        Can be overridden to perform work before stream initialization.
        """

    def connectionInitialized(self):
        """
        The XML stream has been initialized.

        Can be overridden to perform work after stream initialization, e.g. to
        set up observers and start exchanging XML stanzas.
        """

    def connectionLost(self, reason):
        """
        The XML stream has been closed.

        This method can be extended to inspect the C{reason} argument and
        act on it.
        """
        self.xmlstream = None

    def send(self, obj):
        """
        Send data over the managed XML stream.

        @note: The stream manager maintains a queue for data sent using this
               method when there is no current initialized XML stream. This
               data is then sent as soon as a new stream has been established
               and initialized. Subsequently, L{connectionInitialized} will be
               called again. If this queueing is not desired, use C{send} on
               C{self.xmlstream}.

        @param obj: data to be sent over the XML stream. This is usually an
                    object providing L{domish.IElement}, or serialized XML. See
                    L{xmlstream.XmlStream} for details.
        """
        self.parent.send(obj)

    def request(self, request):
        """
        Send an IQ request and track the response.

        This passes the request to the parent for sending and response
        tracking.

        @see: L{StreamManager.request}.
        """
        return self.parent.request(request)


class StreamManager(XMPPHandlerCollection):
    """
    Business logic representing a managed XMPP connection.

    This maintains a single XMPP connection and provides facilities for packet
    routing and transmission. Business logic modules are objects providing
    L{IXMPPHandler} (like subclasses of L{XMPPHandler}), and added
    using L{addHandler}.

    @ivar xmlstream: currently managed XML stream
    @type xmlstream: L{XmlStream}
    @ivar logTraffic: if true, log all traffic.
    @type logTraffic: L{bool}
    @ivar _initialized: Whether the stream represented by L{xmlstream} has
                        been initialized. This is used when caching outgoing
                        stanzas.
    @type _initialized: L{bool}
    @ivar _packetQueue: internal buffer of unsent data. See L{send} for details.
    @type _packetQueue: L{list}
    @ivar timeout: Default IQ request timeout in seconds.
    @type timeout: L{int}
    @ivar _reactor: A provider of L{IReactorTime} to track timeouts.
    """

    timeout = None
    _reactor = None

    logTraffic = False

    def __init__(self, factory, reactor=None):
        """
        Construct a stream manager.

        @param factory: The stream factory to connect with.
        @param reactor: A provider of L{IReactorTime} to track timeouts.
            If not provided, the global reactor will be used.
        """
        XMPPHandlerCollection.__init__(self)
        self.xmlstream = None
        self._packetQueue = []
        self._initialized = False

        factory.addBootstrap(xmlstream.STREAM_CONNECTED_EVENT, self._connected)
        factory.addBootstrap(xmlstream.STREAM_AUTHD_EVENT, self._authd)
        factory.addBootstrap(xmlstream.INIT_FAILED_EVENT, self.initializationFailed)
        factory.addBootstrap(xmlstream.STREAM_END_EVENT, self._disconnected)
        self.factory = factory

        if reactor is None:
            from twisted.internet import reactor
        self._reactor = reactor

        # Set up IQ response tracking
        self._iqDeferreds = {}

    def addHandler(self, handler):
        """
        Add protocol handler.

        When an XML stream has already been established, the handler's
        C{connectionInitialized} will be called to get it up to speed.
        """
        XMPPHandlerCollection.addHandler(self, handler)

        # get protocol handler up to speed when a connection has already
        # been established
        if self.xmlstream:
            handler.makeConnection(self.xmlstream)
        if self._initialized:
            handler.connectionInitialized()

    def _connected(self, xs):
        """
        Called when the transport connection has been established.

        Here we optionally set up traffic logging (depending on L{logTraffic})
        and call each handler's C{makeConnection} method with the L{XmlStream}
        instance.
        """

        def logDataIn(buf):
            log.msg("RECV: %r" % buf)

        def logDataOut(buf):
            log.msg("SEND: %r" % buf)

        if self.logTraffic:
            xs.rawDataInFn = logDataIn
            xs.rawDataOutFn = logDataOut

        self.xmlstream = xs

        for e in list(self):
            e.makeConnection(xs)

    def _authd(self, xs):
        """
        Called when the stream has been initialized.

        Send out cached stanzas and call each handler's
        C{connectionInitialized} method.
        """

        xs.addObserver('/iq[@type="result"]', self._onIQResponse)
        xs.addObserver('/iq[@type="error"]', self._onIQResponse)

        # Flush all pending packets
        for p in self._packetQueue:
            xs.send(p)
        self._packetQueue = []
        self._initialized = True

        # Notify all child services which implement
        # the IService interface
        for e in list(self):
            e.connectionInitialized()

    def initializationFailed(self, reason):
        """
        Called when stream initialization has failed.

        Stream initialization has halted, with the reason indicated by
        C{reason}. It may be retried by calling the authenticator's
        C{initializeStream}. See the respective authenticators for details.

        @param reason: A failure instance indicating why stream initialization
                       failed.
        @type reason: L{failure.Failure}
        """

    def _disconnected(self, reason):
        """
        Called when the stream has been closed.

        From this point on, the manager doesn't interact with the
        L{XmlStream} anymore and notifies each handler that the connection
        was lost by calling its C{connectionLost} method.
        """
        self.xmlstream = None
        self._initialized = False

        # Twisted versions before 11.0 passed an XmlStream here.
        if not hasattr(reason, "trap"):
            reason = failure.Failure(ConnectionDone())

        # Notify all child services which implement
        # the IService interface
        for e in list(self):
            e.connectionLost(reason)

        # This errbacks all deferreds of iq's for which no response has
        # been received with a L{ConnectionLost} failure. Otherwise, the
        # deferreds will never be fired.
        iqDeferreds = self._iqDeferreds
        self._iqDeferreds = {}
        for d in iqDeferreds.values():
            d.errback(reason)

    def _onIQResponse(self, iq):
        """
        Handle iq response by firing associated deferred.
        """
        try:
            d = self._iqDeferreds[iq["id"]]
        except KeyError:
            return

        del self._iqDeferreds[iq["id"]]
        iq.handled = True
        if iq["type"] == "error":
            d.errback(error.exceptionFromStanza(iq))
        else:
            d.callback(iq)

    def send(self, obj):
        """
        Send data over the XML stream.

        When there is no established XML stream, the data is queued and sent
        out when a new XML stream has been established and initialized.

        @param obj: data to be sent over the XML stream. See
                    L{xmlstream.XmlStream.send} for details.
        """
        if self._initialized:
            self.xmlstream.send(obj)
        else:
            self._packetQueue.append(obj)

    def request(self, request):
        """
        Send an IQ request and track the response.

        A request is an IQ L{generic.Stanza} of type C{'get'} or C{'set'}. It
        will have its C{toElement} called to render to a
        L{Element<twisted.words.xish.domish.Element>} which is then sent out
        over the current stream. If there is no such stream (yet), it is queued
        and sent whenever a connection is established and initialized, just
        like L{send}.

        If the request doesn't have an identifier, it will be assigned a fresh
        one, so the response can be tracked.

        The deferred that is returned will fire with the
        L{Element<twisted.words.xish.domish.Element>} representation of the
        response if it is a result iq. If the response is an error iq, a
        corresponding L{error.StanzaError} will be errbacked.

        If the connection is closed before a response was received, the deferred
        will be errbacked with the reason failure.

        A request may also have a timeout, either by setting a default timeout
        in L{StreamManager}'s C{timeout} attribute or on the C{timeout}
        attribute of the request.

        @param request: The IQ request.
        @type request: L{generic.Request}
        """
        if request.stanzaKind != "iq" or request.stanzaType not in ("get", "set"):
            return defer.fail(ValueError("Not a request"))

        element = request.toElement()

        # Make sure we have a trackable id on the stanza
        if not request.stanzaID:
            element.addUniqueId()
            request.stanzaID = element["id"]

        # Set up iq response tracking
        d = defer.Deferred()
        self._iqDeferreds[element["id"]] = d

        timeout = getattr(request, "timeout", self.timeout)

        if timeout is not None:

            def onTimeout():
                del self._iqDeferreds[element["id"]]
                d.errback(xmlstream.TimeoutError("IQ timed out"))

            call = self._reactor.callLater(timeout, onTimeout)

            def cancelTimeout(result):
                if call.active():
                    call.cancel()

                return result

            d.addBoth(cancelTimeout)
        self.send(element)
        return d


class IQHandlerMixin(object):
    """
    XMPP subprotocol mixin for handle incoming IQ stanzas.

    This matches up the iq with XPath queries to call methods on itself,
    wrapping the call so that exceptions result in proper error responses,
    or, when succesful will reply with a response with optional payload.

    Derivatives of this class must provide an
    L{XmlStream<twisted.words.protocols.jabber.xmlstream.XmlStream>} instance
    in its C{xmlstream} attribute.

    The optional payload is taken from the result of the handler and is
    expected to be a child or a list of childs.

    If an exception is raised, or the deferred has its errback called,
    the exception is checked for being a L{error.StanzaError}. If so,
    an error response is sent. Any other exception will cause a error
    response of C{internal-server-error} to be sent.

    A typical way to use this mixin, is to set up L{xpath} observers on the
    C{xmlstream} to call handleRequest, for example in an overridden
    L{XMPPHandler.connectionMade}. It is likely a good idea to only listen for
    incoming iq get and/org iq set requests, and not for any iq, to prevent
    hijacking incoming responses to outgoing iq requests. An example:

        >>> QUERY_ROSTER = "/query[@xmlns='jabber:iq:roster']"
        >>> class MyHandler(XMPPHandler, IQHandlerMixin):
        ...    iqHandlers = {"/iq[@type='get']" + QUERY_ROSTER: 'onRosterGet',
        ...                  "/iq[@type='set']" + QUERY_ROSTER: 'onRosterSet'}
        ...    def connectionMade(self):
        ...        self.xmlstream.addObserver(
        ...          "/iq[@type='get' or @type='set']" + QUERY_ROSTER,
        ...          self.handleRequest)
        ...    def onRosterGet(self, iq):
        ...        pass
        ...    def onRosterSet(self, iq):
        ...        pass

    @cvar iqHandlers: Mapping from XPath queries (as a string) to the method
                      name that will handle requests that match the query.
    @type iqHandlers: L{dict}
    """

    iqHandlers = None

    def handleRequest(self, iq):
        """
        Find a handler and wrap the call for sending a response stanza.
        """

        def toResult(result, iq):
            response = toResponse(iq, "result")

            if result:
                if IElement.providedBy(result):
                    response.addChild(result)
                else:
                    for element in result:
                        response.addChild(element)

            return response

        def checkNotImplemented(failure):
            failure.trap(NotImplementedError)
            raise error.StanzaError("feature-not-implemented")

        def fromStanzaError(failure, iq):
            failure.trap(error.StanzaError)
            return failure.value.toResponse(iq)

        def fromOtherError(failure, iq):
            log.msg("Unhandled error in iq handler:", isError=True)
            log.err(failure)
            return error.StanzaError("internal-server-error").toResponse(iq)

        handler = None
        for queryString, method in self.iqHandlers.items():
            if xpath.internQuery(queryString).matches(iq):
                handler = getattr(self, method)

        if handler:
            d = defer.maybeDeferred(handler, iq)
        else:
            d = defer.fail(NotImplementedError())

        d.addCallback(toResult, iq)
        d.addErrback(checkNotImplemented)
        d.addErrback(fromStanzaError, iq)
        d.addErrback(fromOtherError, iq)

        d.addCallback(self.send)

        iq.handled = True


__all__ = ["XMPPHandler", "StreamManager", "IQHandlerMixin"]



# -*- coding: utf-8 -*-
# -*- test-case-name: wokkel.test.test_rsm -*-
#
# Copyright (c) Adrien Cossa, 2016
# Copyright (c) Jérôme Poisson, 2019-2026
# See LICENSE for details.

"""
XMPP Result Set Management protocol.

This protocol is specified in
U{XEP-0059<http://xmpp.org/extensions/xep-0059.html>}.
"""

from twisted.words.xish import domish
from twisted.words.protocols.jabber import error
from wokkel import disco

from . import pubsub
import copy

NS_RSM = "http://jabber.org/protocol/rsm"
NS_PUBSUB_RSM = "http://jabber.org/protocol/pubsub#rsm"


class RSMError(error.StanzaError):
    """
    RSM error.
    """

    def __init__(self, text=None):
        error.StanzaError.__init__(self, "bad-request", text=text)


class RSMNotFoundError(Exception):
    """
    An expected RSM element has not been found.
    """


class RSMRequest(object):
    """
    A Result Set Management request.

    @ivar max_: limit on the number of retrieved items.
    @itype max_: C{int} or C{unicode}

    @ivar index: starting index of the requested page.
    @itype index: C{int} or C{unicode} or C{None}

    @ivar after: ID of the element immediately preceding the page.
    @itype after: C{unicode}

    @ivar before: ID of the element immediately following the page.
    @itype before: C{unicode}
    """

    def __init__(self, max_=10, after=None, before=None, index=None):
        self.max = int(max_)

        if index is not None:
            assert after is None and before is None
            index = int(index)
        self.index = index

        if after is not None:
            assert before is None
            assert isinstance(after, str)
        self.after = after

        if before is not None:
            assert isinstance(before, str)
        self.before = before

    def __str__(self):
        return "RSM Request: max={0.max} after={0.after} before={0.before} index={0.index}".format(
            self
        )

    @classmethod
    def fromElement(cls, element):
        """Parse the given request element.

        @param element: request containing a set element, or set element itself.
        @type element: L{domish.Element}

        @return: RSMRequest instance.
        @rtype: L{RSMRequest}
        """

        if element.name == "set" and element.uri == NS_RSM:
            set_elt = element
        else:
            try:
                set_elt = next(element.elements(NS_RSM, "set"))
            except StopIteration:
                raise RSMNotFoundError()

        try:
            before_elt = next(set_elt.elements(NS_RSM, "before"))
        except StopIteration:
            before = None
        else:
            before = str(before_elt)

        try:
            after_elt = next(set_elt.elements(NS_RSM, "after"))
        except StopIteration:
            after = None
        else:
            after = str(after_elt)
            if not after:
                raise RSMError("<after/> element can't be empty in RSM request")

        try:
            max_elt = next(set_elt.elements(NS_RSM, "max"))
        except StopIteration:
            # FIXME: even if it doesn't make a lot of sense without it
            #        <max/> element is not mandatory in XEP-0059
            raise RSMError("RSM request is missing its 'max' element")
        else:
            try:
                max_ = int(str(max_elt))
            except ValueError:
                raise RSMError("bad value for 'max' element")

        try:
            index_elt = next(set_elt.elements(NS_RSM, "index"))
        except StopIteration:
            index = None
        else:
            try:
                index = int(str(index_elt))
            except ValueError:
                raise RSMError("bad value for 'index' element")

        return RSMRequest(max_, after, before, index)

    def toElement(self):
        """
        Return the DOM representation of this RSM request.

        @rtype: L{domish.Element}
        """
        set_elt = domish.Element((NS_RSM, "set"))
        set_elt.addElement("max", content=str(self.max))

        if self.index is not None:
            set_elt.addElement("index", content=str(self.index))

        if self.before is not None:
            if self.before == "":  # request the last page
                set_elt.addElement("before")
            else:
                set_elt.addElement("before", content=self.before)

        if self.after is not None:
            set_elt.addElement("after", content=self.after)

        return set_elt

    def render(self, element):
        """Embed the DOM representation of this RSM request in the given element.

        @param element: Element to contain the RSM request.
        @type element: L{domish.Element}

        @return: RSM request element.
        @rtype: L{domish.Element}
        """
        set_elt = self.toElement()
        element.addChild(set_elt)

        return set_elt


class RSMResponse(object):
    """
    A Result Set Management response.

    @ivar first: ID of the first element of the returned page.
    @itype first: C{unicode}

    @ivar last: ID of the last element of the returned page.
    @itype last: C{unicode}

    @ivar index: starting index of the returned page.
    @itype index: C{int}

    @ivar count: total number of items.
    @itype count: C{int}

    """

    def __init__(self, first=None, last=None, index=None, count=None):
        if first is None:
            assert last is None and index is None
        if last is None:
            assert first is None
        self.first = first
        self.last = last
        if count is not None:
            self.count = int(count)
        else:
            self.count = None
        if index is not None:
            self.index = int(index)
        else:
            self.index = None

    def __str__(self):
        return "RSM Request: first={0.first} last={0.last} index={0.index} count={0.count}".format(
            self
        )

    @classmethod
    def fromElement(cls, element):
        """Parse the given response element.

        @param element: response element.
        @type element: L{domish.Element}

        @return: RSMResponse instance.
        @rtype: L{RSMResponse}
        """
        try:
            set_elt = next(element.elements(NS_RSM, "set"))
        except StopIteration:
            raise RSMNotFoundError()

        try:
            first_elt = next(set_elt.elements(NS_RSM, "first"))
        except StopIteration:
            first = None
            index = None
        else:
            first = str(first_elt)
            try:
                index = int(first_elt["index"])
            except KeyError:
                index = None
            except ValueError:
                raise RSMError("bad index in RSM response")

        try:
            last_elt = next(set_elt.elements(NS_RSM, "last"))
        except StopIteration:
            if first is not None:
                raise RSMError("RSM response is missing its 'last' element")
            else:
                last = None
        else:
            if first is None:
                raise RSMError("RSM response is missing its 'first' element")
            last = str(last_elt)

        try:
            count_elt = next(set_elt.elements(NS_RSM, "count"))
        except StopIteration:
            count = None
        else:
            try:
                count = int(str(count_elt))
            except ValueError:
                raise RSMError("invalid count in RSM response")

        return RSMResponse(first, last, index, count)

    def toElement(self):
        """
        Return the DOM representation of this RSM request.

        @rtype: L{domish.Element}
        """
        set_elt = domish.Element((NS_RSM, "set"))
        if self.first is not None:
            first_elt = set_elt.addElement("first", content=self.first)
            if self.index is not None:
                first_elt["index"] = str(self.index)

            set_elt.addElement("last", content=self.last)

        if self.count is not None:
            set_elt.addElement("count", content=str(self.count))

        return set_elt

    def render(self, element):
        """Embed the DOM representation of this RSM response in the given element.

        @param element: Element to contain the RSM response.
        @type element:  L{domish.Element}

        @return: RSM request element.
        @rtype: L{domish.Element}
        """
        set_elt = self.toElement()
        element.addChild(set_elt)
        return set_elt

    def toDict(self):
        """Return a dict representation of the object.

        @return: a dict of strings.
        @rtype: C{dict} binding C{unicode} to C{unicode}
        """
        result = {}
        for attr in ("first", "last", "index", "count"):
            value = getattr(self, attr)
            if value is not None:
                result[attr] = value
        return result


class PubSubRequest(pubsub.PubSubRequest):
    """PubSubRequest extension to handle RSM.

    @ivar rsm: RSM request instance.
    @type rsm: L{RSMRequest}
    """

    rsm = None
    _parameters = copy.deepcopy(pubsub.PubSubRequest._parameters)
    _parameters["items"].append("rsm")

    def _parse_rsm(self, verbElement):
        try:
            self.rsm = RSMRequest.fromElement(verbElement.parent)
        except RSMNotFoundError:
            self.rsm = None

    def _render_rsm(self, verbElement):
        if self.rsm:
            self.rsm.render(verbElement.parent)


class PubSubClient(pubsub.PubSubClient):
    """PubSubClient extension to handle RSM."""

    _request_class = PubSubRequest

    def items(
        self,
        service,
        nodeIdentifier,
        maxItems=None,
        subscriptionIdentifier=None,
        sender=None,
        itemIdentifiers=None,
        orderBy=None,
        rsm_request=None,
    ):
        """
        Retrieve previously published items from a publish subscribe node.

        @param service: The publish subscribe service that keeps the node.
        @type service: L{JID<twisted.words.protocols.jabber.jid.JID>}

        @param nodeIdentifier: The identifier of the node.
        @type nodeIdentifier: C{unicode}

        @param maxItems: Optional limit on the number of retrieved items.
        @type maxItems: C{int}

        @param subscriptionIdentifier: Optional subscription identifier. In
            case the node has been subscribed to multiple times, this narrows
            the results to the specific subscription.
        @type subscriptionIdentifier: C{unicode}

        @param sender: Optional sender address.
        @type sender: L{JID<twisted.words.protocols.jabber.jid.JID>}

        @param itemIdentifiers: Identifiers of the items to be retrieved.
        @type itemIdentifiers: C{set}

        @param orderBy: Keys to order by
        @type orderBy: L{list} of L{unicode}

        @param rsm_request: RSM request instance.
        @type rsm_request: L{RSMRequest}

        @return: a Deferred that fires a C{list} of C{tuple} of L{domish.Element}, L{RSMResponse}.
        @rtype: L{defer.Deferred}
        """
        # XXX: we have to copy initial method instead of calling it,
        #      as original cb remove all non item elements
        request = self._request_class("items")
        request.recipient = service
        request.nodeIdentifier = nodeIdentifier
        if maxItems:
            request.maxItems = str(int(maxItems))
        request.subscriptionIdentifier = subscriptionIdentifier
        request.sender = sender
        request.itemIdentifiers = itemIdentifiers
        request.orderBy = orderBy
        request.rsm = rsm_request

        def cb(iq):
            items = []
            pubsub_elt = iq.pubsub
            if pubsub_elt.items:
                for element in pubsub_elt.items.elements(pubsub.NS_PUBSUB, "item"):
                    items.append(element)

            try:
                rsm_response = RSMResponse.fromElement(pubsub_elt)
            except RSMNotFoundError:
                rsm_response = None
            return (items, rsm_response)

        d = request.send(self.xmlstream)
        d.addCallback(cb)
        return d


class PubSubService(pubsub.PubSubService):
    """PubSubService extension to handle RSM."""

    _request_class = PubSubRequest

    def _toResponse_items(self, resp_tuple, resource, request):
        elts, rsm_response = resp_tuple

        response = pubsub.PubSubService._toResponse_items(self, elts, resource, request)
        if rsm_response is not None:
            response.addChild(rsm_response.toElement())

        return response

    def getDiscoInfo(self, requestor, target, nodeIdentifier=""):
        def appendRSMFeatures(info):
            for ns in (NS_RSM, NS_PUBSUB_RSM):
                feature = disco.DiscoFeature(ns)
                if feature not in info:
                    info.append(feature)
            return info

        d = super().getDiscoInfo(requestor, target, nodeIdentifier)
        d.addCallback(appendRSMFeatures)
        return d


PubSubService._legacyHandlers = copy.deepcopy(pubsub.PubSubService._legacyHandlers)
PubSubService._legacyHandlers["items"][1].append("rsm")



# -*- coding: utf-8 -*-
#
# Copyright (c) Adrien Cossa, Jérôme Poisson
# See LICENSE for details.

"""
Tests for L{wokkel.rsm}.
"""

from zope.interface import verify

from twisted.trial import unittest
from twisted.words.xish import domish
from twisted.words.protocols.jabber.jid import JID
from twisted.words.protocols.jabber.xmlstream import toResponse
from twisted.internet import defer

from wokkel.generic import parseXml
from wokkel import iwokkel, pubsub
from wokkel.rsm import NS_RSM, RSMRequest, RSMResponse, PubSubClient, PubSubService
from wokkel.test.helpers import XmlStreamStub, TestableRequestHandlerMixin

import uuid

RSMResponse.__eq__ = (
    lambda self, other: self.first == other.first
    and self.last == other.last
    and self.index == other.index
    and self.count == other.count
)


class RSMRequestTest(unittest.TestCase):
    """
    Tests for L{rsm.RSMRequest}.
    """

    def test___init__(self):
        """
        Fail to initialize a RSMRequest with wrong attribute values.
        """
        self.assertRaises(AssertionError, RSMRequest, index=371, after="test")
        self.assertRaises(AssertionError, RSMRequest, index=371, before="test")
        self.assertRaises(AssertionError, RSMRequest, before=117)
        self.assertRaises(AssertionError, RSMRequest, after=312)
        self.assertRaises(AssertionError, RSMRequest, after="117", before="312")

    def test_parse(self):
        """
        Parse a request element asking for the first page.
        """
        xml = """
        <query xmlns='jabber:iq:search'>
          <nick>Pete</nick>
          <set xmlns='http://jabber.org/protocol/rsm'>
            <max>1</max>
          </set>
        </query>
        """
        request = RSMRequest.fromElement(parseXml(xml))
        self.assertEqual(1, request.max)
        self.assertIdentical(None, request.index)
        self.assertIdentical(None, request.after)
        self.assertIdentical(None, request.before)

    def test_parseSecondPage(self):
        """
        Parse a request element asking for a next page.
        """
        xml = """
        <query xmlns='jabber:iq:search'>
          <nick>Pete</nick>
          <set xmlns='http://jabber.org/protocol/rsm'>
            <max>3</max>
            <after>peterpan@neverland.lit</after>
          </set>
        </query>
        """
        request = RSMRequest.fromElement(parseXml(xml))
        self.assertEqual(3, request.max)
        self.assertIdentical(None, request.index)
        self.assertEqual("peterpan@neverland.lit", request.after)
        self.assertIdentical(None, request.before)

    def test_parsePreviousPage(self):
        """
        Parse a request element asking for a previous page.
        """
        xml = """
        <query xmlns='jabber:iq:search'>
          <nick>Pete</nick>
          <set xmlns='http://jabber.org/protocol/rsm'>
            <max>5</max>
            <before>peterpan@pixyland.org</before>
          </set>
        </query>
        """
        request = RSMRequest.fromElement(parseXml(xml))
        self.assertEqual(5, request.max)
        self.assertIdentical(None, request.index)
        self.assertIdentical(None, request.after)
        self.assertEqual("peterpan@pixyland.org", request.before)

    def test_parseLastPage(self):
        """
        Parse a request element asking for the last page.
        """
        xml = """
        <query xmlns='jabber:iq:search'>
          <nick>Pete</nick>
          <set xmlns='http://jabber.org/protocol/rsm'>
            <max>7</max>
            <before/>
          </set>
        </query>
        """
        request = RSMRequest.fromElement(parseXml(xml))
        self.assertEqual(7, request.max)
        self.assertIdentical(None, request.index)
        self.assertIdentical(None, request.after)
        self.assertEqual("", request.before)

    def test_parseOutOfOrderPage(self):
        """
        Parse a request element asking for a page out of order.
        """
        xml = """
        <query xmlns='jabber:iq:search'>
          <nick>Pete</nick>
          <set xmlns='http://jabber.org/protocol/rsm'>
            <max>9</max>
            <index>371</index>
          </set>
        </query>
        """
        request = RSMRequest.fromElement(parseXml(xml))
        self.assertEqual(9, request.max)
        self.assertEqual(371, request.index)
        self.assertIdentical(None, request.after)
        self.assertIdentical(None, request.before)

    def test_parseItemCount(self):
        """
        Parse a request element asking for the items count.
        """
        xml = """
        <query xmlns='jabber:iq:search'>
          <nick>Pete</nick>
          <set xmlns='http://jabber.org/protocol/rsm'>
            <max>0</max>
          </set>
        </query>
        """
        request = RSMRequest.fromElement(parseXml(xml))
        self.assertEqual(0, request.max)
        self.assertIdentical(None, request.index)
        self.assertIdentical(None, request.after)
        self.assertIdentical(None, request.before)

    def test_render(self):
        """
        Embed a page request in the element.
        """
        element = domish.Element(("jabber:iq:search", "query"))
        element.addElement("items")["max_items"] = "10"
        RSMRequest(1).render(element)

        self.assertEqual("10", element.items["max_items"])  # not changed

        self.assertEqual(NS_RSM, element.set.uri)
        self.assertEqual("1", "".join(element.set.max.children))
        self.assertIdentical(None, element.set.after)
        self.assertIdentical(None, element.set.before)
        self.assertIdentical(None, element.set.index)

    def test_renderPubSub(self):
        """
        Embed a page request in the pubsub element.
        """
        element = domish.Element((pubsub.NS_PUBSUB, "pubsub"))
        element.addElement("items")["max_items"] = "10"
        RSMRequest(3).render(element)

        self.assertEqual("10", element.items["max_items"])  # not changed

        self.assertEqual(NS_RSM, element.set.uri)
        self.assertEqual("3", "".join(element.set.max.children))
        self.assertIdentical(None, element.set.after)
        self.assertIdentical(None, element.set.before)
        self.assertIdentical(None, element.set.index)

    def test_renderItems(self):
        """
        Embed a page request in the element, specify items.
        """
        element = domish.Element(("jabber:iq:search", "query"))
        RSMRequest(5, index=127).render(element)
        self.assertEqual(NS_RSM, element.set.uri)
        self.assertEqual("5", "".join(element.set.max.children))
        self.assertIdentical(None, element.set.after)
        self.assertIdentical(None, element.set.before)
        self.assertEqual("127", "".join(element.set.index.children))

    def test_renderAfter(self):
        """
        Embed a page request in the element, specify after.
        """
        element = domish.Element(("jabber:iq:search", "query"))
        RSMRequest(5, after="test").render(element)
        self.assertEqual(NS_RSM, element.set.uri)
        self.assertEqual("5", "".join(element.set.max.children))
        self.assertEqual("test", "".join(element.set.after.children))
        self.assertIdentical(None, element.set.before)
        self.assertIdentical(None, element.set.index)

    def test_renderBefore(self):
        """
        Embed a page request in the element, specify before.
        """
        element = domish.Element(("jabber:iq:search", "query"))
        RSMRequest(5, before="test").render(element)
        self.assertEqual(NS_RSM, element.set.uri)
        self.assertEqual("5", "".join(element.set.max.children))
        self.assertIdentical(None, element.set.after)
        self.assertEqual("test", "".join(element.set.before.children))
        self.assertIdentical(None, element.set.index)


class RSMResponseTest(unittest.TestCase):
    """
    Tests for L{rsm.RSMResponse}.
    """

    def test___init__(self):
        """
        Fail to initialize a RSMResponse with wrong attribute values.
        """
        self.assertRaises(AssertionError, RSMResponse, index=127, first="127")
        self.assertRaises(AssertionError, RSMResponse, index=127, last="351")

    def test_parse(self):
        """
        Parse a response element returning a page.
        """
        xml = """
        <query xmlns='jabber:iq:search'>
          <set xmlns='http://jabber.org/protocol/rsm'>
            <first index='20'>stpeter@jabber.org</first>
            <last>peterpan@neverland.lit</last>
            <count>800</count>
          </set>
        </query>
        """
        response = RSMResponse.fromElement(parseXml(xml))
        self.assertEqual(800, response.count)
        self.assertEqual(20, response.index)
        self.assertEqual("stpeter@jabber.org", response.first)
        self.assertEqual("peterpan@neverland.lit", response.last)

    def test_parseEmptySet(self):
        """
        Parse a response element returning an empty set.
        """
        xml = """
        <query xmlns='jabber:iq:search'>
          <set xmlns='http://jabber.org/protocol/rsm'>
            <count>800</count>
          </set>
        </query>
        """
        response = RSMResponse.fromElement(parseXml(xml))
        self.assertEqual(800, response.count)
        self.assertIdentical(None, response.first)
        self.assertIdentical(None, response.last)
        self.assertIdentical(None, response.index)

    def test_render(self):
        """
        Embed a page response in the element.
        """
        element = domish.Element(("jabber:iq:search", "query"))
        RSMResponse("stpeter@jabber.org", "peterpan@neverland.lit", 20, 800).render(
            element
        )

        self.assertEqual(NS_RSM, element.set.uri)
        self.assertEqual("800", "".join(element.set.count.children))
        self.assertEqual("stpeter@jabber.org", "".join(element.set.first.children))
        self.assertEqual("peterpan@neverland.lit", "".join(element.set.last.children))
        self.assertEqual("20", element.set.first["index"])

    def test_renderEmptySet(self):
        """
        Embed a page response in the element, for empty set.
        """
        element = domish.Element(("jabber:iq:search", "query"))
        RSMResponse(count=800).render(element)

        self.assertEqual(NS_RSM, element.set.uri)
        self.assertEqual("800", "".join(element.set.count.children))
        self.assertIdentical(None, element.set.first)
        self.assertIdentical(None, element.set.last)


class PubSubClientTest(unittest.TestCase):
    """
    Tests for L{rsm.PubSubClient}.
    """

    timeout = 2

    def setUp(self):
        self.stub = XmlStreamStub()
        self.protocol = PubSubClient()
        self.protocol.xmlstream = self.stub.xmlstream
        self.protocol.connectionInitialized()

    def test_items(self):
        """
        Test sending items request to get the first page.
        """

        def cb(response):
            items, rsm = response
            self.assertEquals(2, len(items))
            self.assertEquals([item1, item2], items)
            self.assertEquals(rsm, RSMResponse("item1", "item2", 0, 800))

        d = self.protocol.items(
            JID("pubsub.example.org"), "test", rsm_request=RSMRequest(2)
        )
        d.addCallback(cb)

        iq = self.stub.output[-1]
        self.assertEquals("pubsub.example.org", iq.getAttribute("to"))
        self.assertEquals("get", iq.getAttribute("type"))
        self.assertEquals("pubsub", iq.pubsub.name)
        self.assertEquals(pubsub.NS_PUBSUB, iq.pubsub.uri)
        children = list(
            domish.generateElementsQNamed(iq.pubsub.children, "items", pubsub.NS_PUBSUB)
        )
        self.assertEquals(1, len(children))
        child = children[0]
        self.assertEquals("test", child["node"])

        set_elts = list(domish.generateElementsQNamed(iq.pubsub.children, "set", NS_RSM))
        self.assertEquals(1, len(set_elts))
        set_elt = set_elts[0]
        self.assertEquals("2", "".join(set_elt.max.children))

        response = toResponse(iq, "result")
        items = response.addElement((pubsub.NS_PUBSUB, "pubsub")).addElement("items")
        items["node"] = "test"
        item1 = items.addElement("item")
        item1["id"] = "item1"
        item2 = items.addElement("item")
        item2["id"] = "item2"
        RSMResponse("item1", "item2", 0, 800).render(response.pubsub)
        self.stub.send(response)

        return d

    def test_itemsAfter(self):
        """
        Test sending items request to get the next page.
        """

        def cb(response):
            items, rsm = response
            self.assertEquals(2, len(items))
            self.assertEquals([item1, item2], items)
            self.assertEquals(rsm, RSMResponse("item3", "item4", 2, 800))

        d = self.protocol.items(
            JID("pubsub.example.org"), "test", rsm_request=RSMRequest(2, after="item2")
        )
        d.addCallback(cb)

        iq = self.stub.output[-1]
        self.assertEquals("pubsub.example.org", iq.getAttribute("to"))
        self.assertEquals("get", iq.getAttribute("type"))
        self.assertEquals("pubsub", iq.pubsub.name)
        self.assertEquals(pubsub.NS_PUBSUB, iq.pubsub.uri)
        children = list(
            domish.generateElementsQNamed(iq.pubsub.children, "items", pubsub.NS_PUBSUB)
        )
        self.assertEquals(1, len(children))
        child = children[0]
        self.assertEquals("test", child["node"])

        set_elts = list(domish.generateElementsQNamed(iq.pubsub.children, "set", NS_RSM))
        self.assertEquals(1, len(set_elts))
        set_elt = set_elts[0]
        self.assertEquals("2", "".join(set_elt.max.children))
        self.assertEquals("item2", "".join(set_elt.after.children))

        response = toResponse(iq, "result")
        items = response.addElement((pubsub.NS_PUBSUB, "pubsub")).addElement("items")
        items["node"] = "test"
        item1 = items.addElement("item")
        item1["id"] = "item3"
        item2 = items.addElement("item")
        item2["id"] = "item4"
        RSMResponse("item3", "item4", 2, 800).render(response.pubsub)
        self.stub.send(response)

        return d

    def test_itemsBefore(self):
        """
        Test sending items request to get the previous page.
        """

        def cb(response):
            items, rsm = response
            self.assertEquals(2, len(items))
            self.assertEquals([item1, item2], items)
            self.assertEquals(rsm, RSMResponse("item1", "item2", 0, 800))

        d = self.protocol.items(
            JID("pubsub.example.org"), "test", rsm_request=RSMRequest(2, before="item3")
        )
        d.addCallback(cb)

        iq = self.stub.output[-1]
        self.assertEquals("pubsub.example.org", iq.getAttribute("to"))
        self.assertEquals("get", iq.getAttribute("type"))
        self.assertEquals("pubsub", iq.pubsub.name)
        self.assertEquals(pubsub.NS_PUBSUB, iq.pubsub.uri)
        children = list(
            domish.generateElementsQNamed(iq.pubsub.children, "items", pubsub.NS_PUBSUB)
        )
        self.assertEquals(1, len(children))
        child = children[0]
        self.assertEquals("test", child["node"])

        set_elts = list(domish.generateElementsQNamed(iq.pubsub.children, "set", NS_RSM))
        self.assertEquals(1, len(set_elts))
        set_elt = set_elts[0]
        self.assertEquals("2", "".join(set_elt.max.children))
        self.assertEquals("item3", "".join(set_elt.before.children))

        response = toResponse(iq, "result")
        items = response.addElement((pubsub.NS_PUBSUB, "pubsub")).addElement("items")
        items["node"] = "test"
        item1 = items.addElement("item")
        item1["id"] = "item1"
        item2 = items.addElement("item")
        item2["id"] = "item2"
        RSMResponse("item1", "item2", 0, 800).render(response.pubsub)
        self.stub.send(response)

        return d

    def test_itemsIndex(self):
        """
        Test sending items request to get a page out of order.
        """

        def cb(response):
            items, rsm = response
            self.assertEquals(3, len(items))
            self.assertEquals([item1, item2, item3], items)
            self.assertEquals(rsm, RSMResponse("item4", "item6", 3, 800))

        d = self.protocol.items(
            JID("pubsub.example.org"), "test", rsm_request=RSMRequest(3, index=3)
        )
        d.addCallback(cb)

        iq = self.stub.output[-1]
        self.assertEquals("pubsub.example.org", iq.getAttribute("to"))
        self.assertEquals("get", iq.getAttribute("type"))
        self.assertEquals("pubsub", iq.pubsub.name)
        self.assertEquals(pubsub.NS_PUBSUB, iq.pubsub.uri)
        children = list(
            domish.generateElementsQNamed(iq.pubsub.children, "items", pubsub.NS_PUBSUB)
        )
        self.assertEquals(1, len(children))
        child = children[0]
        self.assertEquals("test", child["node"])

        set_elts = list(domish.generateElementsQNamed(iq.pubsub.children, "set", NS_RSM))
        self.assertEquals(1, len(set_elts))
        set_elt = set_elts[0]
        self.assertEquals("3", "".join(set_elt.max.children))
        self.assertEquals("3", "".join(set_elt.index.children))

        response = toResponse(iq, "result")
        items = response.addElement((pubsub.NS_PUBSUB, "pubsub")).addElement("items")
        items["node"] = "test"
        item1 = items.addElement("item")
        item1["id"] = "item4"
        item2 = items.addElement("item")
        item2["id"] = "item5"
        item3 = items.addElement("item")
        item3["id"] = "item6"
        RSMResponse("item4", "item6", 3, 800).render(response.pubsub)
        self.stub.send(response)

        return d

    def test_itemsCount(self):
        """
        Test sending items request to count them.
        """

        def cb(response):
            items, rsm = response
            self.assertEquals(0, len(items))
            self.assertEquals(rsm, RSMResponse(count=800))

        d = self.protocol.items(
            JID("pubsub.example.org"), "test", rsm_request=RSMRequest(0)
        )
        d.addCallback(cb)

        iq = self.stub.output[-1]
        self.assertEquals("pubsub.example.org", iq.getAttribute("to"))
        self.assertEquals("get", iq.getAttribute("type"))
        self.assertEquals("pubsub", iq.pubsub.name)
        self.assertEquals(pubsub.NS_PUBSUB, iq.pubsub.uri)
        children = list(
            domish.generateElementsQNamed(iq.pubsub.children, "items", pubsub.NS_PUBSUB)
        )
        self.assertEquals(1, len(children))
        child = children[0]
        self.assertEquals("test", child["node"])

        set_elts = list(domish.generateElementsQNamed(iq.pubsub.children, "set", NS_RSM))
        self.assertEquals(1, len(set_elts))
        set_elt = set_elts[0]
        self.assertEquals("0", "".join(set_elt.max.children))

        response = toResponse(iq, "result")
        response.addElement((pubsub.NS_PUBSUB, "pubsub"))
        RSMResponse(count=800).render(response.pubsub)
        self.stub.send(response)

        return d


class PubSubServiceTest(unittest.TestCase, TestableRequestHandlerMixin):

    def setUp(self):
        self.stub = XmlStreamStub()
        self.resource = pubsub.PubSubResource()
        self.service = PubSubService(self.resource)
        self.service.send = self.stub.xmlstream.send

    def test_on_items(self):
        """
        On a items request, return the first item for the given node.
        """
        xml = """
        <iq type='get' to='pubsub.example.org'
                       from='user@example.org'>
          <pubsub xmlns='http://jabber.org/protocol/pubsub'>
            <items node='test'/>
          </pubsub>
          <set xmlns='http://jabber.org/protocol/rsm'>
            <max>1</max>
          </set>
        </iq>
        """

        def items(request):
            rsm = RSMResponse("item", "item", 0, 800)
            return defer.succeed(([pubsub.Item("current")], rsm))

        def cb(element):
            self.assertEqual(pubsub.NS_PUBSUB, element.uri)
            self.assertEqual(pubsub.NS_PUBSUB, element.items.uri)
            self.assertEqual(1, len(element.items.children))
            item = element.items.children[-1]
            self.assertTrue(domish.IElement.providedBy(item))
            self.assertEqual("item", item.name)
            self.assertEqual(pubsub.NS_PUBSUB, item.uri)
            self.assertEqual("current", item["id"])
            self.assertEqual(NS_RSM, element.set.uri)
            self.assertEqual("800", "".join(element.set.count.children))
            self.assertEqual("0", element.set.first["index"])
            self.assertEqual("item", "".join(element.set.first.children))
            self.assertEqual("item", "".join(element.set.last.children))

        self.resource.items = items
        verify.verifyObject(iwokkel.IPubSubResource, self.resource)
        d = self.handleRequest(xml)
        d.addCallback(cb)
        return d

    def test_on_itemsIndex(self):
        """
        On a items request, return some items out of order for the given node.
        """
        xml = """
        <iq type='get' to='pubsub.example.org'
                       from='user@example.org'>
          <pubsub xmlns='http://jabber.org/protocol/pubsub'>
            <items node='test'/>
          </pubsub>
          <set xmlns='http://jabber.org/protocol/rsm'>
            <max>2</max>
            <index>3</index>
          </set>
        </iq>
        """

        def items(request):
            rsm = RSMResponse("i1", "i2", 3, 800)
            return defer.succeed(([pubsub.Item("i1"), pubsub.Item("i2")], rsm))

        def cb(element):
            self.assertEqual(pubsub.NS_PUBSUB, element.uri)
            self.assertEqual(pubsub.NS_PUBSUB, element.items.uri)
            self.assertEqual(2, len(element.items.children))
            item = element.items.children[0]
            self.assertTrue(domish.IElement.providedBy(item))
            self.assertEqual("item", item.name)
            self.assertEqual(pubsub.NS_PUBSUB, item.uri)
            self.assertEqual("i1", item["id"])
            item = element.items.children[1]
            self.assertTrue(domish.IElement.providedBy(item))
            self.assertEqual("item", item.name)
            self.assertEqual(pubsub.NS_PUBSUB, item.uri)
            self.assertEqual("i2", item["id"])
            self.assertEqual(NS_RSM, element.set.uri)
            self.assertEqual("800", "".join(element.set.count.children))
            self.assertEqual("3", element.set.first["index"])
            self.assertEqual("i1", "".join(element.set.first.children))
            self.assertEqual("i2", "".join(element.set.last.children))

        self.resource.items = items
        verify.verifyObject(iwokkel.IPubSubResource, self.resource)
        d = self.handleRequest(xml)
        d.addCallback(cb)
        return d

    def test_on_itemsCount(self):
        """
        On a items request, return the items count.
        """
        xml = """
        <iq type='get' to='pubsub.example.org'
                       from='user@example.org'>
          <pubsub xmlns='http://jabber.org/protocol/pubsub'>
            <items node='test'/>
          </pubsub>
          <set xmlns='http://jabber.org/protocol/rsm'>
            <max>0</max>
          </set>
        </iq>
        """

        def items(request):
            rsm = RSMResponse(count=800)
            return defer.succeed(([], rsm))

        def cb(element):
            self.assertEqual(pubsub.NS_PUBSUB, element.uri)
            self.assertEqual(pubsub.NS_PUBSUB, element.items.uri)
            self.assertEqual(0, len(element.items.children))
            self.assertEqual(NS_RSM, element.set.uri)
            self.assertEqual("800", "".join(element.set.count.children))
            self.assertEqual(None, element.set.first)
            self.assertEqual(None, element.set.last)

        self.resource.items = items
        verify.verifyObject(iwokkel.IPubSubResource, self.resource)
        d = self.handleRequest(xml)
        d.addCallback(cb)
        return d



# -*- test-case-name: wokkel.test.test_compat -*-
#
# Copyright (c) Twisted Matrix Laboratories.
# See LICENSE for details.

"""
Compatibility module to provide backwards compatibility with Twisted features.
"""

from __future__ import division, absolute_import

from twisted.words.protocols.jabber import xmlstream


class IQ(xmlstream.IQ):
    def __init__(self, *args, **kwargs):
        # Make sure we have a reactor parameter
        try:
            reactor = kwargs["reactor"]
        except KeyError:
            from twisted.internet import reactor
        kwargs["reactor"] = reactor

        # Check if IQ's init accepts the reactor parameter
        try:
            xmlstream.IQ.__init__(self, *args, **kwargs)
        except TypeError:
            # Guess not. Remove the reactor parameter and try again.
            del kwargs["reactor"]
            xmlstream.IQ.__init__(self, *args, **kwargs)

            # Patch the XmlStream instance so that it has a _callLater
            self._xmlstream._callLater = reactor.callLater


__all__ = ["IQ"]



# Copyright (c) Ralph Meijer.
# See LICENSE for details.

from twisted.application.service import ServiceMaker

WokkelXMPPComponentServer = ServiceMaker(
    "XMPP Component Server",
    "wokkel.componentservertap",
    "An XMPP Component Server",
    "wokkel-component-server",
)



# Copyright (c) Ralph Meijer.
# See LICENSE for details.

"""
Tests for L{wokkel.server}.
"""

from __future__ import division, absolute_import
from twisted.internet import defer
from twisted.python import failure

try:
    from twisted.internet.testing import StringTransport
except ImportError:
    from twisted.test.proto_helpers import StringTransport

from twisted.trial import unittest
from twisted.words.protocols.jabber import error, jid, xmlstream
from twisted.words.xish import domish

from wokkel import component, server

NS_STREAMS = "http://etherx.jabber.org/streams"
NS_DIALBACK = "jabber:server:dialback"


class GenerateKeyTest(unittest.TestCase):
    """
    Tests for L{server.generateKey}.
    """

    def testBasic(self):
        originating = "example.org"
        receiving = "xmpp.example.com"
        sid = "D60000229F"
        secret = "s3cr3tf0rd14lb4ck"

        key = server.generateKey(secret, receiving, originating, sid)

        self.assertEqual(
            key, "37c69b1cf07a3f67c04a5ef5902fa5114f2c76fe4a2686482ba5b89323075643"
        )


class XMPPServerListenAuthenticatorTest(unittest.TestCase):
    """
    Tests for L{server.XMPPServerListenAuthenticator}.
    """

    secret = "s3cr3tf0rd14lb4ck"
    originating = "example.org"
    receiving = "xmpp.example.com"
    sid = "D60000229F"
    key = "37c69b1cf07a3f67c04a5ef5902fa5114f2c76fe4a2686482ba5b89323075643"

    def setUp(self):
        self.output = []

        class MyService(object):
            pass

        self.service = MyService()
        self.service.defaultDomain = self.receiving
        self.service.domains = [self.receiving, "pubsub." + self.receiving]
        self.service.secret = self.secret

        self.authenticator = server.XMPPServerListenAuthenticator(self.service)
        self.xmlstream = xmlstream.XmlStream(self.authenticator)
        self.xmlstream.send = self.output.append
        self.xmlstream.transport = StringTransport()

    def test_attributes(self):
        """
        Test attributes of authenticator and stream objects.
        """
        self.assertEqual(self.service, self.authenticator.service)
        self.assertEqual(self.xmlstream.initiating, False)

    def test_streamStartedVersion0(self):
        """
        The authenticator supports pre-XMPP 1.0 streams.
        """
        self.xmlstream.connectionMade()
        self.xmlstream.dataReceived(
            "<stream:stream xmlns:stream='http://etherx.jabber.org/streams' "
            "xmlns:db='jabber:server:dialback' "
            "xmlns='jabber:server' "
            "to='xmpp.example.com'>"
        )
        self.assertEqual((0, 0), self.xmlstream.version)

    def test_streamStartedVersion1(self):
        """
        The authenticator supports XMPP 1.0 streams.
        """
        self.xmlstream.connectionMade()
        self.xmlstream.dataReceived(
            "<stream:stream xmlns:stream='http://etherx.jabber.org/streams' "
            "xmlns:db='jabber:server:dialback' "
            "xmlns='jabber:server' "
            "to='xmpp.example.com' "
            "version='1.0'>"
        )
        self.assertEqual((1, 0), self.xmlstream.version)

    def test_streamStartedSID(self):
        """
        The response stream will have a stream ID.
        """
        self.xmlstream.connectionMade()
        self.assertIdentical(None, self.xmlstream.sid)

        self.xmlstream.dataReceived(
            "<stream:stream xmlns:stream='http://etherx.jabber.org/streams' "
            "xmlns:db='jabber:server:dialback' "
            "xmlns='jabber:server' "
            "to='xmpp.example.com' "
            "version='1.0'>"
        )
        self.assertNotIdentical(None, self.xmlstream.sid)

    def test_streamStartedSentResponseHeader(self):
        """
        A stream header is sent in response to the incoming stream header.
        """
        self.xmlstream.connectionMade()
        self.assertFalse(self.xmlstream._headerSent)

        self.xmlstream.dataReceived(
            "<stream:stream xmlns:stream='http://etherx.jabber.org/streams' "
            "xmlns:db='jabber:server:dialback' "
            "xmlns='jabber:server' "
            "to='xmpp.example.com'>"
        )
        self.assertTrue(self.xmlstream._headerSent)

    def test_streamStartedNotSentFeatures(self):
        """
        No features are sent in response to an XMPP < 1.0 stream header.
        """
        self.xmlstream.connectionMade()
        self.xmlstream.dataReceived(
            "<stream:stream xmlns:stream='http://etherx.jabber.org/streams' "
            "xmlns:db='jabber:server:dialback' "
            "xmlns='jabber:server' "
            "to='xmpp.example.com'>"
        )
        self.assertEqual(1, len(self.output))

    def test_streamStartedSentFeatures(self):
        """
        Features are sent in response to an XMPP >= 1.0 stream header.
        """
        self.xmlstream.connectionMade()
        self.xmlstream.dataReceived(
            "<stream:stream xmlns:stream='http://etherx.jabber.org/streams' "
            "xmlns:db='jabber:server:dialback' "
            "xmlns='jabber:server' "
            "to='xmpp.example.com' "
            "version='1.0'>"
        )
        self.assertEqual(2, len(self.output))
        features = self.output[-1]
        self.assertEqual(NS_STREAMS, features.uri)
        self.assertEqual("features", features.name)

    def test_streamRootElement(self):
        """
        Test stream error on wrong stream namespace.
        """
        self.xmlstream.connectionMade()
        self.xmlstream.dataReceived(
            "<stream:stream xmlns:stream='badns' "
            "xmlns:db='jabber:server:dialback' "
            "xmlns='jabber:server' "
            "to='xmpp.example.com'>"
        )

        self.assertEqual(3, len(self.output))
        exc = error.exceptionFromStreamError(self.output[1])
        self.assertEqual("invalid-namespace", exc.condition)

    def test_streamDefaultNamespace(self):
        """
        Test stream error on missing dialback namespace.
        """
        self.xmlstream.connectionMade()
        self.xmlstream.dataReceived(
            "<stream:stream xmlns:stream='http://etherx.jabber.org/streams' "
            "xmlns:db='jabber:server:dialback' "
            "xmlns='badns' "
            "to='xmpp.example.com'>"
        )

        self.assertEqual(3, len(self.output))
        exc = error.exceptionFromStreamError(self.output[1])
        self.assertEqual("invalid-namespace", exc.condition)

    def test_streamNoDialbackNamespace(self):
        """
        Test stream error on missing dialback namespace.
        """
        self.xmlstream.connectionMade()
        self.xmlstream.dataReceived(
            "<stream:stream xmlns:stream='http://etherx.jabber.org/streams' "
            "xmlns='jabber:server' "
            "to='xmpp.example.com'>"
        )

        self.assertEqual(3, len(self.output))
        exc = error.exceptionFromStreamError(self.output[1])
        self.assertEqual("invalid-namespace", exc.condition)

    def test_streamBadDialbackNamespace(self):
        """
        Test stream error on missing dialback namespace.
        """
        self.xmlstream.connectionMade()
        self.xmlstream.dataReceived(
            "<stream:stream xmlns:stream='http://etherx.jabber.org/streams' "
            "xmlns:db='badns' "
            "xmlns='jabber:server' "
            "to='xmpp.example.com'>"
        )

        self.assertEqual(3, len(self.output))
        exc = error.exceptionFromStreamError(self.output[1])
        self.assertEqual("invalid-namespace", exc.condition)

    def test_streamToUnknownHost(self):
        """
        Test stream error on stream's to attribute having unknown host.
        """
        self.xmlstream.connectionMade()
        self.xmlstream.dataReceived(
            "<stream:stream xmlns:stream='http://etherx.jabber.org/streams' "
            "xmlns:db='jabber:server:dialback' "
            "xmlns='jabber:server' "
            "to='badhost'>"
        )

        self.assertEqual(3, len(self.output))
        exc = error.exceptionFromStreamError(self.output[1])
        self.assertEqual("host-unknown", exc.condition)

    def test_streamToOtherLocalHost(self):
        """
        The authenticator supports XMPP 1.0 streams.
        """
        self.xmlstream.connectionMade()
        self.xmlstream.dataReceived(
            "<stream:stream xmlns:stream='http://etherx.jabber.org/streams' "
            "xmlns:db='jabber:server:dialback' "
            "xmlns='jabber:server' "
            "to='pubsub.xmpp.example.com' "
            "version='1.0'>"
        )

        self.assertEqual(2, len(self.output))
        self.assertEqual(jid.JID("pubsub.xmpp.example.com"), self.xmlstream.thisEntity)

    def test_onResult(self):
        def cb(result):
            self.assertEqual(1, len(self.output))
            reply = self.output[0]
            self.assertEqual(self.originating, reply["to"])
            self.assertEqual(self.receiving, reply["from"])
            self.assertEqual("valid", reply["type"])

        def validateConnection(thisHost, otherHost, sid, key):
            self.assertEqual(thisHost, self.receiving)
            self.assertEqual(otherHost, self.originating)
            self.assertEqual(sid, self.sid)
            self.assertEqual(key, self.key)
            return defer.succeed(None)

        self.xmlstream.sid = self.sid
        self.service.validateConnection = validateConnection

        result = domish.Element((NS_DIALBACK, "result"))
        result["to"] = self.receiving
        result["from"] = self.originating
        result.addContent(self.key)

        d = self.authenticator.onResult(result)
        d.addCallback(cb)
        return d

    def test_onResultFailure(self):
        class TestError(Exception):
            pass

        def cb(result):
            reply = self.output[0]
            self.assertEqual("invalid", reply["type"])
            self.assertEqual(1, len(self.flushLoggedErrors(TestError)))

        def validateConnection(thisHost, otherHost, sid, key):
            return defer.fail(TestError())

        self.xmlstream.sid = self.sid
        self.service.validateConnection = validateConnection

        result = domish.Element((NS_DIALBACK, "result"))
        result["to"] = self.receiving
        result["from"] = self.originating
        result.addContent(self.key)

        d = self.authenticator.onResult(result)
        d.addCallback(cb)
        return d


class FakeService(object):
    domains = set(["example.org", "pubsub.example.org"])
    defaultDomain = "example.org"
    secret = "mysecret"

    def __init__(self):
        self.dispatched = []

    def dispatch(self, xs, element):
        self.dispatched.append(element)


class XMPPS2SServerFactoryTest(unittest.TestCase):
    """
    Tests for L{component.XMPPS2SServerFactory}.
    """

    def setUp(self):
        self.service = FakeService()
        self.factory = server.XMPPS2SServerFactory(self.service)
        self.xmlstream = self.factory.buildProtocol(None)
        self.transport = StringTransport()
        self.xmlstream.thisEntity = jid.JID("example.org")
        self.xmlstream.otherEntity = jid.JID("example.com")

    def test_makeConnection(self):
        """
        A new connection increases the stream serial count. No logs by default.
        """
        self.xmlstream.makeConnection(self.transport)
        self.assertEqual(0, self.xmlstream.serial)
        self.assertEqual(1, self.factory.serial)
        self.assertIdentical(None, self.xmlstream.rawDataInFn)
        self.assertIdentical(None, self.xmlstream.rawDataOutFn)

    def test_makeConnectionLogTraffic(self):
        """
        Setting logTraffic should set up raw data loggers.
        """
        self.factory.logTraffic = True
        self.xmlstream.makeConnection(self.transport)
        self.assertNotIdentical(None, self.xmlstream.rawDataInFn)
        self.assertNotIdentical(None, self.xmlstream.rawDataOutFn)

    def test_onError(self):
        """
        An observer for stream errors should trigger onError to log it.
        """
        self.xmlstream.makeConnection(self.transport)

        class TestError(Exception):
            pass

        reason = failure.Failure(TestError())
        self.xmlstream.dispatch(reason, xmlstream.STREAM_ERROR_EVENT)
        self.assertEqual(1, len(self.flushLoggedErrors(TestError)))

    def test_connectionInitialized(self):
        """ """
        self.xmlstream.makeConnection(self.transport)
        self.xmlstream.dispatch(self.xmlstream, xmlstream.STREAM_AUTHD_EVENT)

    def test_connectionLost(self):
        """ """
        self.xmlstream.makeConnection(self.transport)
        self.xmlstream.dispatch(self.xmlstream, xmlstream.STREAM_AUTHD_EVENT)
        self.xmlstream.dispatch(None, xmlstream.STREAM_END_EVENT)

    def test_Element(self):
        self.xmlstream.makeConnection(self.transport)
        self.xmlstream.dispatch(self.xmlstream, xmlstream.STREAM_AUTHD_EVENT)

        stanza = domish.Element((None, "presence"))
        self.xmlstream.dispatch(stanza)
        self.assertEqual(1, len(self.service.dispatched))
        self.assertIdentical(stanza, self.service.dispatched[-1])

    def test_ElementNotAuthenticated(self):
        self.xmlstream.makeConnection(self.transport)

        stanza = domish.Element((None, "presence"))
        self.xmlstream.dispatch(stanza)
        self.assertEqual(0, len(self.service.dispatched))


class ServerServiceTest(unittest.TestCase):

    def setUp(self):
        self.output = []

        self.xmlstream = xmlstream.XmlStream(xmlstream.Authenticator())
        self.xmlstream.thisEntity = jid.JID("example.org")
        self.xmlstream.otherEntity = jid.JID("example.com")
        self.xmlstream.send = self.output.append

        self.router = component.Router()
        self.service = server.ServerService(
            self.router, secret="mysecret", domain="example.org"
        )
        self.service.xmlstream = self.xmlstream

    def test_defaultDomainInDomains(self):
        """
        The default domain is part of the domains considered local.
        """
        self.assertIn(self.service.defaultDomain, self.service.domains)

    def test_dispatch(self):
        stanza = domish.Element((None, "presence"))
        stanza["to"] = "user@example.org"
        stanza["from"] = "other@example.com"
        self.service.dispatch(self.xmlstream, stanza)

        self.assertEqual(1, len(self.output))
        self.assertIdentical(stanza, self.output[-1])

    def test_dispatchNoTo(self):
        errors = []
        self.xmlstream.sendStreamError = errors.append

        stanza = domish.Element((None, "presence"))
        stanza["from"] = "other@example.com"
        self.service.dispatch(self.xmlstream, stanza)

        self.assertEqual(1, len(errors))

    def test_generatedSecret(self):
        self.router = component.Router()
        self.service = server.ServerService(self.router, domain="example.org")
        self.assertEqual(32, len(self.service.secret))



# -*- test-case-name: wokkel.test.test_shim -*-
#
# Copyright (c) Ralph Meijer.
# See LICENSE for details.

"""
XMPP Stanza Headers and Internet Metadata.

This protocol is specified in
U{XEP-0131<http://xmpp.org/extensions/xep-0131.html>}.
"""

from __future__ import division, absolute_import

from twisted.words.xish import domish

NS_SHIM = "http://jabber.org/protocol/shim"


class Headers(domish.Element):
    def __init__(self, headers):
        domish.Element.__init__(self, (NS_SHIM, "headers"))
        for name, value in headers:
            self.addElement("header", content=value)["name"] = name


def extractHeaders(stanza):
    """
    Extract SHIM headers from stanza.

    @param stanza: The stanza to extract headers from.
    @type stanza: L{Element<twisted.words.xish.domish.Element>}
    @return: Headers as a mapping from header name to a list of values.
    @rtype: L{dict}
    """
    headers = {}

    for element in domish.generateElementsQNamed(stanza.children, "headers", NS_SHIM):
        for header in domish.generateElementsQNamed(element.children, "header", NS_SHIM):
            headers.setdefault(header["name"], []).append(str(header))

    return headers



# Copyright (c) Ralph Meijer.
# See LICENSE for details.

"""
Tests for L{wokkel.ping}.
"""

from __future__ import division, absolute_import

from zope.interface import verify

from twisted.internet import defer
from twisted.trial import unittest
from twisted.words.protocols.jabber.error import StanzaError
from twisted.words.protocols.jabber.jid import JID
from twisted.words.protocols.jabber.xmlstream import toResponse

from wokkel import disco, iwokkel, ping
from wokkel.generic import parseXml
from wokkel.test.helpers import XmlStreamStub


class PingClientProtocolTest(unittest.TestCase):
    """
    Tests for L{ping.PingClientProtocol}.
    """

    def setUp(self):
        self.stub = XmlStreamStub()
        self.protocol = ping.PingClientProtocol()
        self.protocol.xmlstream = self.stub.xmlstream
        self.protocol.connectionInitialized()

    def test_ping(self):
        """
        Pinging a service should fire a deferred with None
        """

        def cb(result):
            self.assertIdentical(None, result)

        d = self.protocol.ping(JID("example.com"))
        d.addCallback(cb)

        iq = self.stub.output[-1]
        self.assertEqual("example.com", iq.getAttribute("to"))
        self.assertEqual("get", iq.getAttribute("type"))
        self.assertEqual("urn:xmpp:ping", iq.ping.uri)

        response = toResponse(iq, "result")
        self.stub.send(response)

        return d

    def test_pingWithSender(self):
        """
        Pinging a service with a sender address should include that address.
        """
        d = self.protocol.ping(JID("example.com"), sender=JID("user@example.com"))

        iq = self.stub.output[-1]
        self.assertEqual("user@example.com", iq.getAttribute("from"))

        response = toResponse(iq, "result")
        self.stub.send(response)

        return d

    def test_pingNotSupported(self):
        """
        Pinging a service should fire a deferred with None if not supported.
        """

        def cb(result):
            self.assertIdentical(None, result)

        d = self.protocol.ping(JID("example.com"))
        d.addCallback(cb)

        iq = self.stub.output[-1]

        exc = StanzaError("service-unavailable")
        response = exc.toResponse(iq)
        self.stub.send(response)

        return d

    def test_pingStanzaError(self):
        """
        Pinging a service should errback a deferred on other (stanza) errors.
        """

        def cb(exc):
            self.assertEquals("item-not-found", exc.condition)

        d = self.protocol.ping(JID("example.com"))
        self.assertFailure(d, StanzaError)
        d.addCallback(cb)

        iq = self.stub.output[-1]

        exc = StanzaError("item-not-found")
        response = exc.toResponse(iq)
        self.stub.send(response)

        return d


class PingHandlerTest(unittest.TestCase):
    """
    Tests for L{ping.PingHandler}.
    """

    def setUp(self):
        self.stub = XmlStreamStub()
        self.protocol = ping.PingHandler()
        self.protocol.xmlstream = self.stub.xmlstream
        self.protocol.connectionInitialized()

    def test_onPing(self):
        """
        A ping should have a simple result response.
        """
        xml = """<iq from='test@example.com' to='example.com' type='get'>
                   <ping xmlns='urn:xmpp:ping'/>
                 </iq>"""
        self.stub.send(parseXml(xml))

        response = self.stub.output[-1]
        self.assertEquals("example.com", response.getAttribute("from"))
        self.assertEquals("test@example.com", response.getAttribute("to"))
        self.assertEquals("result", response.getAttribute("type"))

    def test_onPingHandled(self):
        """
        The ping handler should mark the stanza as handled.
        """
        xml = """<iq from='test@example.com' to='example.com' type='get'>
                   <ping xmlns='urn:xmpp:ping'/>
                 </iq>"""
        iq = parseXml(xml)
        self.stub.send(iq)

        self.assertTrue(iq.handled)

    def test_interfaceIDisco(self):
        """
        The ping handler should provice Service Discovery information.
        """
        verify.verifyObject(iwokkel.IDisco, self.protocol)

    def test_getDiscoInfo(self):
        """
        The ping namespace should be returned as a supported feature.
        """

        def cb(info):
            discoInfo = disco.DiscoInfo()
            for item in info:
                discoInfo.append(item)
            self.assertIn("urn:xmpp:ping", discoInfo.features)

        d = defer.maybeDeferred(
            self.protocol.getDiscoInfo,
            JID("user@example.org/home"),
            JID("pubsub.example.org"),
            "",
        )
        d.addCallback(cb)
        return d

    def test_getDiscoInfoNode(self):
        """
        The ping namespace should not be returned for a node.
        """

        def cb(info):
            discoInfo = disco.DiscoInfo()
            for item in info:
                discoInfo.append(item)
            self.assertNotIn("urn:xmpp:ping", discoInfo.features)

        d = defer.maybeDeferred(
            self.protocol.getDiscoInfo,
            JID("user@example.org/home"),
            JID("pubsub.example.org"),
            "test",
        )
        d.addCallback(cb)
        return d

    def test_getDiscoItems(self):
        """
        Items are not supported by this handler, so an empty list is expected.
        """

        def cb(items):
            self.assertEquals(0, len(items))

        d = defer.maybeDeferred(
            self.protocol.getDiscoItems,
            JID("user@example.org/home"),
            JID("pubsub.example.org"),
            "",
        )
        d.addCallback(cb)
        return d



# Copyright (c) Ralph Meijer.
# See LICENSE for details

"""
Wokkel.

Support library for Twisted applications using XMPP protocols.
"""

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("tx-xmpp")
except PackageNotFoundError:
    __version__ = "0.0.0"

__all__ = ["__version__"]



# -*- test-case-name: wokkel.test.test_delay -*-
#
# Copyright (c) Ralph Meijer.
# See LICENSE for details.

"""
Delayed Delivery.

Support for comunicating Delayed Delivery information as specified by
U{XEP-0203<http://xmpp.org/extensions/xep-0203.html>} and its predecessor
U{XEP-0091<http://xmpp.org/extensions/xep-0091.html>}.
"""

from __future__ import division, absolute_import

from dateutil.parser import parse
from dateutil.tz import tzutc

from twisted.words.protocols.jabber.jid import InvalidFormat, JID
from twisted.words.xish import domish

NS_DELAY = "urn:xmpp:delay"
NS_JABBER_DELAY = "jabber:x:delay"


class Delay(object):
    """
    Delayed Delivery information.

    Instances of this class represent delayed delivery information that can be
    parsed from and rendered into both XEP-0203 and legacy XEP-0091 formats.

    @ivar stamp: The timestamp the stanza was originally sent.
    @type stamp: L{datetime.datetime}
    @ivar sender: The optional entity that originally sent the stanza or
        delayed its delivery.
    @type sender: L{JID}
    """

    def __init__(self, stamp, sender=None):
        self.stamp = stamp
        self.sender = sender

    def toElement(self, legacy=False):
        """
        Render this instance into a domish Element.

        @param legacy: If C{True}, use the legacy XEP-0091 format.
        @type legacy: L{bool}
        """
        if not self.stamp:
            raise ValueError("stamp is required")
        if self.stamp.tzinfo is None:
            raise ValueError("stamp is not offset-aware")

        if legacy:
            element = domish.Element((NS_JABBER_DELAY, "x"))
            stampFormat = "%Y%m%dT%H:%M:%S"
        else:
            element = domish.Element((NS_DELAY, "delay"))
            stampFormat = "%Y-%m-%dT%H:%M:%SZ"

        stamp = self.stamp.astimezone(tzutc())
        element["stamp"] = stamp.strftime(stampFormat)

        if self.sender:
            element["from"] = self.sender.full()

        return element

    @staticmethod
    def fromElement(element):
        """
        Create an instance from a domish Element.
        """
        try:
            stamp = parse(element["stamp"])

            # Assume UTC if no timezone was given
            if stamp.tzinfo is None:
                stamp = stamp.replace(tzinfo=tzutc())
        except (KeyError, ValueError, TypeError):
            stamp = None

        try:
            sender = JID(element["from"])
        except (KeyError, InvalidFormat):
            sender = None

        delay = Delay(stamp, sender)
        return delay


class DelayMixin(object):
    """
    Mixin for parsing delayed delivery information from stanzas.

    This can be used as a mixin for subclasses of L{wokkel.generic.Stanza}
    for parsing delayed delivery information. If both XEP-0203 and XEP-0091
    formats are present, the former takes precedence.
    """

    delay = None

    childParsers = {
        (NS_DELAY, "delay"): "_childParser_delay",
        (NS_JABBER_DELAY, "x"): "_childParser_legacyDelay",
    }

    def _childParser_delay(self, element):
        self.delay = Delay.fromElement(element)

    def _childParser_legacyDelay(self, element):
        if not self.delay:
            self.delay = Delay.fromElement(element)



# Copyright (c) Ralph Meijer.
# See LICENSE for details.

"""
Tests for L{wokkel.subprotocols}
"""

from __future__ import division, absolute_import

from zope.interface.verify import verifyObject

from twisted.trial import unittest
from twisted.test import proto_helpers
from twisted.internet import defer, task
from twisted.internet.error import ConnectionDone
from twisted.python import failure
from twisted.words.xish import domish
from twisted.words.protocols.jabber import error, ijabber, xmlstream

from wokkel import generic, subprotocols


class DummyFactory(object):
    """
    Dummy XmlStream factory that only registers bootstrap observers.
    """

    def __init__(self):
        self.callbacks = {}

    def addBootstrap(self, event, callback):
        self.callbacks[event] = callback


class DummyXMPPHandler(subprotocols.XMPPHandler):
    """
    Dummy XMPP subprotocol handler to count the methods are called on it.
    """

    def __init__(self):
        self.doneMade = 0
        self.doneInitialized = 0
        self.doneLost = 0

    def makeConnection(self, xs):
        self.connectionMade()

    def connectionMade(self):
        self.doneMade += 1

    def connectionInitialized(self):
        self.doneInitialized += 1

    def connectionLost(self, reason):
        self.doneLost += 1


class FailureReasonXMPPHandler(subprotocols.XMPPHandler):
    """
    Dummy handler specifically for failure Reason tests.
    """

    def __init__(self):
        self.gotFailureReason = False

    def connectionLost(self, reason):
        if isinstance(reason, failure.Failure):
            self.gotFailureReason = True


class IQGetStanza(generic.Stanza):
    timeout = None

    stanzaKind = "iq"
    stanzaType = "get"
    stanzaID = "test"


class XMPPHandlerTest(unittest.TestCase):
    """
    Tests for L{subprotocols.XMPPHandler}.
    """

    def test_interface(self):
        """
        L{xmlstream.XMPPHandler} implements L{ijabber.IXMPPHandler}.
        """
        verifyObject(ijabber.IXMPPHandler, subprotocols.XMPPHandler())

    def test_send(self):
        """
        Test that data is passed on for sending by the stream manager.
        """

        class DummyStreamManager(object):
            def __init__(self):
                self.outlist = []

            def send(self, data):
                self.outlist.append(data)

        handler = subprotocols.XMPPHandler()
        handler.parent = DummyStreamManager()
        handler.send("<presence/>")
        self.assertEquals(["<presence/>"], handler.parent.outlist)

    def test_makeConnection(self):
        """
        Test that makeConnection saves the XML stream and calls connectionMade.
        """

        class TestXMPPHandler(subprotocols.XMPPHandler):
            def connectionMade(self):
                self.doneMade = True

        handler = TestXMPPHandler()
        xs = xmlstream.XmlStream(xmlstream.Authenticator())
        handler.makeConnection(xs)
        self.assertTrue(handler.doneMade)
        self.assertIdentical(xs, handler.xmlstream)

    def test_connectionLost(self):
        """
        Test that connectionLost forgets the XML stream.
        """
        handler = subprotocols.XMPPHandler()
        xs = xmlstream.XmlStream(xmlstream.Authenticator())
        handler.makeConnection(xs)
        handler.connectionLost(Exception())
        self.assertIdentical(None, handler.xmlstream)

    def test_request(self):
        """
        A request is passed up to the stream manager.
        """

        class DummyStreamManager(object):
            def __init__(self):
                self.requests = []

            def request(self, request):
                self.requests.append(request)
                return defer.succeed(None)

        handler = subprotocols.XMPPHandler()
        handler.parent = DummyStreamManager()
        request = IQGetStanza()
        d = handler.request(request)
        self.assertEquals(1, len(handler.parent.requests))
        self.assertIdentical(request, handler.parent.requests[-1])
        return d


class StreamManagerTest(unittest.TestCase):
    """
    Tests for L{subprotocols.StreamManager}.
    """

    def setUp(self):
        factory = xmlstream.XmlStreamFactory(xmlstream.Authenticator())
        self.clock = task.Clock()
        self.streamManager = subprotocols.StreamManager(factory, self.clock)
        self.xmlstream = factory.buildProtocol(None)
        self.transport = proto_helpers.StringTransport()
        self.xmlstream.transport = self.transport

        self.request = IQGetStanza()

    def _streamStarted(self):
        """
        Bring the test stream to the initialized state.
        """
        self.xmlstream.connectionMade()
        self.xmlstream.dataReceived(
            b"<stream:stream xmlns='jabber:client' "
            b"xmlns:stream='http://etherx.jabber.org/streams' "
            b"from='example.com' id='12345'>"
        )
        self.xmlstream.dispatch(self.xmlstream, "//event/stream/authd")

    def test_basic(self):
        """
        Test correct initialization and setup of factory observers.
        """
        factory = DummyFactory()
        sm = subprotocols.StreamManager(factory)
        self.assertIdentical(None, sm.xmlstream)
        self.assertEquals([], sm.handlers)
        self.assertEquals(sm._connected, sm.factory.callbacks["//event/stream/connected"])
        self.assertEquals(sm._authd, sm.factory.callbacks["//event/stream/authd"])
        self.assertEquals(sm._disconnected, sm.factory.callbacks["//event/stream/end"])
        self.assertEquals(
            sm.initializationFailed, sm.factory.callbacks["//event/xmpp/initfailed"]
        )

    def test_connected(self):
        """
        Test that protocol handlers have their connectionMade method called
        when the XML stream is connected.
        """
        sm = self.streamManager
        handler = DummyXMPPHandler()
        handler.setHandlerParent(sm)
        xs = xmlstream.XmlStream(xmlstream.Authenticator())
        sm._connected(xs)
        self.assertEquals(1, handler.doneMade)
        self.assertEquals(0, handler.doneInitialized)
        self.assertEquals(0, handler.doneLost)

    def test_connectedLogTrafficFalse(self):
        """
        Test raw data functions unset when logTraffic is set to False.
        """
        sm = self.streamManager
        handler = DummyXMPPHandler()
        handler.setHandlerParent(sm)
        xs = xmlstream.XmlStream(xmlstream.Authenticator())
        sm._connected(xs)
        self.assertIdentical(None, xs.rawDataInFn)
        self.assertIdentical(None, xs.rawDataOutFn)

    def test_connectedLogTrafficTrue(self):
        """
        Test raw data functions set when logTraffic is set to True.
        """
        sm = self.streamManager
        sm.logTraffic = True
        handler = DummyXMPPHandler()
        handler.setHandlerParent(sm)
        xs = xmlstream.XmlStream(xmlstream.Authenticator())
        sm._connected(xs)
        self.assertNotIdentical(None, xs.rawDataInFn)
        self.assertNotIdentical(None, xs.rawDataOutFn)

    def test_authd(self):
        """
        Test that protocol handlers have their connectionInitialized method
        called when the XML stream is initialized.
        """
        sm = self.streamManager
        handler = DummyXMPPHandler()
        handler.setHandlerParent(sm)
        xs = xmlstream.XmlStream(xmlstream.Authenticator())
        sm._authd(xs)
        self.assertEquals(0, handler.doneMade)
        self.assertEquals(1, handler.doneInitialized)
        self.assertEquals(0, handler.doneLost)

    def test_disconnected(self):
        """
        Protocol handlers have connectionLost called on stream disconnect.
        """
        sm = self.streamManager
        handler = DummyXMPPHandler()
        handler.setHandlerParent(sm)
        sm._disconnected(None)
        self.assertEquals(0, handler.doneMade)
        self.assertEquals(0, handler.doneInitialized)
        self.assertEquals(1, handler.doneLost)

    def test_disconnectedReason(self):
        """
        A L{STREAM_END_EVENT} results in L{StreamManager} firing the handlers
        L{connectionLost} methods, passing a L{failure.Failure} reason.
        """
        sm = self.streamManager
        handler = FailureReasonXMPPHandler()
        handler.setHandlerParent(sm)
        xmlstream.XmlStream(xmlstream.Authenticator())
        sm._disconnected(failure.Failure(Exception("no reason")))
        self.assertEquals(True, handler.gotFailureReason)

    def test_addHandler(self):
        """
        Test the addition of a protocol handler while not connected.
        """
        sm = self.streamManager
        handler = DummyXMPPHandler()
        handler.setHandlerParent(sm)

        self.assertEquals(0, handler.doneMade)
        self.assertEquals(0, handler.doneInitialized)
        self.assertEquals(0, handler.doneLost)

    def test_addHandlerConnected(self):
        """
        Adding a handler when connected doesn't call connectionInitialized.
        """
        sm = self.streamManager
        xs = xmlstream.XmlStream(xmlstream.Authenticator())
        sm._connected(xs)
        handler = DummyXMPPHandler()
        handler.setHandlerParent(sm)

        self.assertEquals(1, handler.doneMade)
        self.assertEquals(0, handler.doneInitialized)
        self.assertEquals(0, handler.doneLost)

    def test_addHandlerConnectedNested(self):
        """
        Adding a handler in connectionMade doesn't cause 2nd call.
        """

        class NestingHandler(DummyXMPPHandler):
            nestedHandler = None

            def connectionMade(self):
                DummyXMPPHandler.connectionMade(self)
                self.nestedHandler = DummyXMPPHandler()
                self.nestedHandler.setHandlerParent(self.parent)

        sm = self.streamManager
        xs = xmlstream.XmlStream(xmlstream.Authenticator())
        handler = NestingHandler()
        handler.setHandlerParent(sm)
        sm._connected(xs)

        self.assertEquals(1, handler.doneMade)
        self.assertEquals(0, handler.doneInitialized)
        self.assertEquals(0, handler.doneLost)

        self.assertEquals(1, handler.nestedHandler.doneMade)
        self.assertEquals(0, handler.nestedHandler.doneInitialized)
        self.assertEquals(0, handler.nestedHandler.doneLost)

    def test_addHandlerInitialized(self):
        """
        Test the addition of a protocol handler after the stream
        have been initialized.

        Make sure that the handler will have the connected stream
        passed via C{makeConnection} and have C{connectionInitialized}
        called.
        """
        sm = self.streamManager
        xs = xmlstream.XmlStream(xmlstream.Authenticator())
        sm._connected(xs)
        sm._authd(xs)
        handler = DummyXMPPHandler()
        handler.setHandlerParent(sm)

        self.assertEquals(1, handler.doneMade)
        self.assertEquals(1, handler.doneInitialized)
        self.assertEquals(0, handler.doneLost)

    def test_addHandlerInitializedNested(self):
        """
        Adding a handler in connectionInitialized doesn't cause 2nd call.
        """

        class NestingHandler(DummyXMPPHandler):
            nestedHandler = None

            def connectionInitialized(self):
                DummyXMPPHandler.connectionInitialized(self)
                self.nestedHandler = DummyXMPPHandler()
                self.nestedHandler.setHandlerParent(self.parent)

        sm = self.streamManager
        xs = xmlstream.XmlStream(xmlstream.Authenticator())
        handler = NestingHandler()
        handler.setHandlerParent(sm)
        sm._connected(xs)
        sm._authd(xs)

        self.assertEquals(1, handler.doneMade)
        self.assertEquals(1, handler.doneInitialized)
        self.assertEquals(0, handler.doneLost)

        self.assertEquals(1, handler.nestedHandler.doneMade)
        self.assertEquals(1, handler.nestedHandler.doneInitialized)
        self.assertEquals(0, handler.nestedHandler.doneLost)

    def test_addHandlerConnectionLostNested(self):
        """
        Adding a handler in connectionLost doesn't call connectionLost there.
        """

        class NestingHandler(DummyXMPPHandler):
            nestedHandler = None

            def connectionLost(self, reason):
                DummyXMPPHandler.connectionLost(self, reason)
                self.nestedHandler = DummyXMPPHandler()
                self.nestedHandler.setHandlerParent(self.parent)

        sm = self.streamManager
        xs = xmlstream.XmlStream(xmlstream.Authenticator())
        handler = NestingHandler()
        handler.setHandlerParent(sm)
        sm._connected(xs)
        sm._authd(xs)
        sm._disconnected(xs)

        self.assertEquals(1, handler.doneMade)
        self.assertEquals(1, handler.doneInitialized)
        self.assertEquals(1, handler.doneLost)

        self.assertEquals(0, handler.nestedHandler.doneMade)
        self.assertEquals(0, handler.nestedHandler.doneInitialized)
        self.assertEquals(0, handler.nestedHandler.doneLost)

    def test_removeHandler(self):
        """
        Test removal of protocol handler.
        """
        sm = self.streamManager
        handler = DummyXMPPHandler()
        handler.setHandlerParent(sm)
        handler.disownHandlerParent(sm)
        self.assertNotIn(handler, sm)
        self.assertIdentical(None, handler.parent)

    def test_sendInitialized(self):
        """
        Test send when the stream has been initialized.

        The data should be sent directly over the XML stream.
        """
        factory = xmlstream.XmlStreamFactory(xmlstream.Authenticator())
        sm = subprotocols.StreamManager(factory)
        xs = factory.buildProtocol(None)
        xs.transport = proto_helpers.StringTransport()
        xs.connectionMade()
        xs.dataReceived(
            b"<stream:stream xmlns='jabber:client' "
            b"xmlns:stream='http://etherx.jabber.org/streams' "
            b"from='example.com' id='12345'>"
        )
        xs.dispatch(xs, "//event/stream/authd")
        sm.send("<presence/>")
        self.assertEquals(b"<presence/>", xs.transport.value())

    def test_sendNotConnected(self):
        """
        Test send when there is no established XML stream.

        The data should be cached until an XML stream has been established and
        initialized.
        """
        factory = xmlstream.XmlStreamFactory(xmlstream.Authenticator())
        sm = subprotocols.StreamManager(factory)
        handler = DummyXMPPHandler()
        sm.addHandler(handler)

        xs = factory.buildProtocol(None)
        xs.transport = proto_helpers.StringTransport()
        sm.send("<presence/>")
        self.assertEquals(b"", xs.transport.value())
        self.assertEquals("<presence/>", sm._packetQueue[0])

        xs.connectionMade()
        self.assertEquals(b"", xs.transport.value())
        self.assertEquals("<presence/>", sm._packetQueue[0])

        xs.dataReceived(
            b"<stream:stream xmlns='jabber:client' "
            b"xmlns:stream='http://etherx.jabber.org/streams' "
            b"from='example.com' id='12345'>"
        )
        xs.dispatch(xs, "//event/stream/authd")

        self.assertEquals(b"<presence/>", xs.transport.value())
        self.assertFalse(sm._packetQueue)

    def test_sendNotInitialized(self):
        """
        Test send when the stream is connected but not yet initialized.

        The data should be cached until the XML stream has been initialized.
        """
        factory = xmlstream.XmlStreamFactory(xmlstream.Authenticator())
        sm = subprotocols.StreamManager(factory)
        xs = factory.buildProtocol(None)
        xs.transport = proto_helpers.StringTransport()
        xs.connectionMade()
        xs.dataReceived(
            b"<stream:stream xmlns='jabber:client' "
            b"xmlns:stream='http://etherx.jabber.org/streams' "
            b"from='example.com' id='12345'>"
        )
        sm.send("<presence/>")
        self.assertEquals(b"", xs.transport.value())
        self.assertEquals("<presence/>", sm._packetQueue[0])

    def test_sendDisconnected(self):
        """
        Test send after XML stream disconnection.

        The data should be cached until a new XML stream has been established
        and initialized.
        """
        factory = xmlstream.XmlStreamFactory(xmlstream.Authenticator())
        sm = subprotocols.StreamManager(factory)
        handler = DummyXMPPHandler()
        sm.addHandler(handler)

        xs = factory.buildProtocol(None)
        xs.connectionMade()
        xs.transport = proto_helpers.StringTransport()
        xs.connectionLost(None)

        sm.send("<presence/>")
        self.assertEquals(b"", xs.transport.value())
        self.assertEquals("<presence/>", sm._packetQueue[0])

    def test_requestSendInitialized(self):
        """
        A request is sent out over the wire when the stream is initialized.
        """
        self._streamStarted()

        self.streamManager.request(self.request)
        self.assertTrue(self.transport.value())

    def test_requestSendInitializedFreshID(self):
        """
        A request without an ID gets a fresh one upon send.
        """
        self._streamStarted()

        self.request.stanzaID = None
        self.streamManager.request(self.request)
        self.assertNotIdentical(None, self.request.stanzaID)
        self.assertTrue(self.transport.value())

    def test_requestSendNotConnected(self):
        """
        A request is queued until a stream is initialized.
        """
        handler = DummyXMPPHandler()
        self.streamManager.addHandler(handler)

        self.streamManager.request(self.request)

        xs = self.xmlstream
        self.assertEquals(b"", xs.transport.value())

        xs.connectionMade()
        self.assertEquals(b"", xs.transport.value())

        xs.dataReceived(
            b"<stream:stream xmlns='jabber:client' "
            b"xmlns:stream='http://etherx.jabber.org/streams' "
            b"from='example.com' id='12345'>"
        )
        xs.dispatch(xs, "//event/stream/authd")

        self.assertTrue(self.transport.value())
        self.assertFalse(self.streamManager._packetQueue)

    def test_requestResultResponse(self):
        """
        A result response gets the request deferred fired with the response.
        """

        def cb(result):
            self.assertEquals(result["type"], "result")

        self._streamStarted()
        d = self.streamManager.request(self.request)
        d.addCallback(cb)

        xs = self.xmlstream
        xs.dataReceived(b"<iq type='result' id='test'/>")
        return d

    def test_requestErrorResponse(self):
        """
        An error response gets the request deferred fired with a failure.
        """
        self._streamStarted()
        d = self.streamManager.request(self.request)
        self.assertFailure(d, error.StanzaError)

        xs = self.xmlstream
        xs.dataReceived(b"<iq type='error' id='test'/>")
        return d

    def test_requestNonTrackedResponse(self):
        """
        Test that untracked iq responses don't trigger any action.

        Untracked means that the id of the incoming response iq is not
        in the stream's C{iqDeferreds} dictionary.
        """
        # Set up a fallback handler that checks the stanza's handled attribute.
        # If that is set to True, the iq tracker claims to have handled the
        # response.
        dispatched = []

        def cb(iq):
            dispatched.append(iq)

        self._streamStarted()
        self.xmlstream.addObserver("/iq", cb, -1)

        # Receive an untracked iq response
        self.xmlstream.dataReceived(b"<iq type='result' id='other'/>")
        self.assertEquals(1, len(dispatched))
        self.assertFalse(getattr(dispatched[-1], "handled", False))

    def test_requestCleanup(self):
        """
        Test if the deferred associated with an iq request is removed
        from the list kept in the L{XmlStream} object after it has
        been fired.
        """
        self._streamStarted()
        d = self.streamManager.request(self.request)
        xs = self.xmlstream
        xs.dataReceived(b"<iq type='result' id='test'/>")
        self.assertNotIn("test", self.streamManager._iqDeferreds)
        return d

    def test_requestDisconnectCleanup(self):
        """
        Test if deferreds for iq's that haven't yet received a response
        have their errback called on stream disconnect.
        """
        d = self.streamManager.request(self.request)
        xs = self.xmlstream
        xs.connectionLost(failure.Failure(ConnectionDone()))
        self.assertFailure(d, ConnectionDone)
        return d

    def test_requestNoModifyingDict(self):
        """
        Test to make sure the errbacks cannot cause the iteration of the
        iqDeferreds to blow up in our face.
        """

        def eb(failure):
            d = xmlstream.IQ(self.xmlstream).send()
            d.addErrback(eb)

        d = self.streamManager.request(self.request)
        d.addErrback(eb)
        self.xmlstream.connectionLost(failure.Failure(ConnectionDone()))
        return d

    def test_requestTimingOut(self):
        """
        Test that an iq request with a defined timeout times out.
        """
        self.request.timeout = 60
        d = self.streamManager.request(self.request)
        self.assertFailure(d, xmlstream.TimeoutError)

        self.clock.pump([1, 60])
        self.assertFalse(self.clock.calls)
        self.assertFalse(self.streamManager._iqDeferreds)
        return d

    def test_requestNotTimingOut(self):
        """
        Test that an iq request with a defined timeout does not time out
        when a response was received before the timeout period elapsed.
        """
        self._streamStarted()
        self.request.timeout = 60
        d = self.streamManager.request(self.request)
        self.clock.callLater(
            1, self.xmlstream.dataReceived, b"<iq type='result' id='test'/>"
        )
        self.clock.pump([1, 1])
        self.assertFalse(self.clock.calls)
        return d

    def test_requestDisconnectTimeoutCancellation(self):
        """
        Test if timeouts for iq's that haven't yet received a response
        are cancelled on stream disconnect.
        """

        self.request.timeout = 60
        d = self.streamManager.request(self.request)

        self.xmlstream.connectionLost(failure.Failure(ConnectionDone()))
        self.assertFailure(d, ConnectionDone)
        self.assertFalse(self.clock.calls)
        return d

    def test_requestNotIQ(self):
        """
        The request stanza must be an iq.
        """
        stanza = generic.Stanza()
        stanza.stanzaKind = "message"

        d = self.streamManager.request(stanza)
        self.assertFailure(d, ValueError)

    def test_requestNotResult(self):
        """
        The request stanza cannot be of type 'result'.
        """
        stanza = generic.Stanza()
        stanza.stanzaKind = "iq"
        stanza.stanzaType = "result"

        d = self.streamManager.request(stanza)
        self.assertFailure(d, ValueError)

    def test_requestNotError(self):
        """
        The request stanza cannot be of type 'error'.
        """
        stanza = generic.Stanza()
        stanza.stanzaKind = "iq"
        stanza.stanzaType = "error"

        d = self.streamManager.request(stanza)
        self.assertFailure(d, ValueError)


class DummyIQHandler(subprotocols.IQHandlerMixin):
    iqHandlers = {'/iq[@type="get"]': "onGet"}

    def __init__(self):
        self.output = []
        self.xmlstream = xmlstream.XmlStream(xmlstream.Authenticator())
        self.xmlstream.send = self.output.append

    def send(self, obj):
        self.xmlstream.send(obj)


class IQHandlerTest(unittest.TestCase):

    def test_match(self):
        """
        Test that the matching handler gets called.
        """

        class Handler(DummyIQHandler):
            called = False

            def onGet(self, iq):
                self.called = True

        iq = domish.Element((None, "iq"))
        iq["type"] = "get"
        iq["id"] = "r1"
        handler = Handler()
        handler.handleRequest(iq)
        self.assertTrue(handler.called)

    def test_noMatch(self):
        """
        Test that the matching handler gets called.
        """

        class Handler(DummyIQHandler):
            called = False

            def onGet(self, iq):
                self.called = True

        iq = domish.Element((None, "iq"))
        iq["type"] = "set"
        iq["id"] = "r1"
        handler = Handler()
        handler.handleRequest(iq)
        self.assertFalse(handler.called)

    def test_success(self):
        """
        Test response when the request is handled successfully.
        """

        class Handler(DummyIQHandler):
            def onGet(self, iq):
                return None

        iq = domish.Element((None, "iq"))
        iq["type"] = "get"
        iq["id"] = "r1"
        handler = Handler()
        handler.handleRequest(iq)
        response = handler.output[-1]
        self.assertEquals(None, response.uri)
        self.assertEquals("iq", response.name)
        self.assertEquals("result", response["type"])

    def test_successPayload(self):
        """
        Test response when the request is handled successfully with payload.
        """

        class Handler(DummyIQHandler):
            payload = domish.Element(("testns", "foo"))

            def onGet(self, iq):
                return self.payload

        iq = domish.Element((None, "iq"))
        iq["type"] = "get"
        iq["id"] = "r1"
        handler = Handler()
        handler.handleRequest(iq)
        response = handler.output[-1]
        self.assertEquals(None, response.uri)
        self.assertEquals("iq", response.name)
        self.assertEquals("result", response["type"])
        payload = next(response.elements())
        self.assertEqual(handler.payload, payload)

    def test_successDeferred(self):
        """
        Test response when where the handler was a deferred.
        """

        class Handler(DummyIQHandler):
            def onGet(self, iq):
                return defer.succeed(None)

        iq = domish.Element((None, "iq"))
        iq["type"] = "get"
        iq["id"] = "r1"
        handler = Handler()
        handler.handleRequest(iq)
        response = handler.output[-1]
        self.assertEquals(None, response.uri)
        self.assertEquals("iq", response.name)
        self.assertEquals("result", response["type"])

    def test_failure(self):
        """
        Test response when the request is handled unsuccessfully.
        """

        class Handler(DummyIQHandler):
            def onGet(self, iq):
                raise error.StanzaError("forbidden")

        iq = domish.Element((None, "iq"))
        iq["type"] = "get"
        iq["id"] = "r1"
        handler = Handler()
        handler.handleRequest(iq)
        response = handler.output[-1]
        self.assertEquals(None, response.uri)
        self.assertEquals("iq", response.name)
        self.assertEquals("error", response["type"])
        e = error.exceptionFromStanza(response)
        self.assertEquals("forbidden", e.condition)

    def test_failureUnknown(self):
        """
        Test response when the request handler raises a non-stanza-error.
        """

        class TestError(Exception):
            pass

        class Handler(DummyIQHandler):
            def onGet(self, iq):
                raise TestError()

        iq = domish.Element((None, "iq"))
        iq["type"] = "get"
        iq["id"] = "r1"
        handler = Handler()
        handler.handleRequest(iq)
        response = handler.output[-1]
        self.assertEquals(None, response.uri)
        self.assertEquals("iq", response.name)
        self.assertEquals("error", response["type"])
        e = error.exceptionFromStanza(response)
        self.assertEquals("internal-server-error", e.condition)
        self.assertEquals(1, len(self.flushLoggedErrors(TestError)))

    def test_notImplemented(self):
        """
        Test response when the request is recognised but not implemented.
        """

        class Handler(DummyIQHandler):
            def onGet(self, iq):
                raise NotImplementedError()

        iq = domish.Element((None, "iq"))
        iq["type"] = "get"
        iq["id"] = "r1"
        handler = Handler()
        handler.handleRequest(iq)
        response = handler.output[-1]
        self.assertEquals(None, response.uri)
        self.assertEquals("iq", response.name)
        self.assertEquals("error", response["type"])
        e = error.exceptionFromStanza(response)
        self.assertEquals("feature-not-implemented", e.condition)

    def test_noHandler(self):
        """
        Test when the request is not recognised.
        """

        iq = domish.Element((None, "iq"))
        iq["type"] = "set"
        iq["id"] = "r1"
        handler = DummyIQHandler()
        handler.handleRequest(iq)
        response = handler.output[-1]
        self.assertEquals(None, response.uri)
        self.assertEquals("iq", response.name)
        self.assertEquals("error", response["type"])
        e = error.exceptionFromStanza(response)
        self.assertEquals("feature-not-implemented", e.condition)



# -*- test-case-name: wokkel.test.test_formats -*-
#
# Copyright (c) Ralph Meijer.
# See LICENSE for details.

"""
Generic payload formats.
"""

from __future__ import division, absolute_import

NS_MOOD = "http://jabber.org/protocol/mood"
NS_TUNE = "http://jabber.org/protocol/tune"


class Mood:
    """
    User mood.

    This represents a user's mood, as defined in
    U{XEP-0107<http://xmpp.org/extensions/xep-0107.html>}.

    @ivar value: The mood value.
    @ivar text: The optional natural-language description of, or reason
                for the mood.
    """

    def __init__(self, value, text=None):
        self.value = value
        self.text = text

    def fromXml(self, element):
        """
        Get a Mood instance from an XML representation.

        This class method parses the given XML document into a L{Mood}
        instances.

        @param element: The XML mood document.
        @type element: object providing
                       L{IElement<twisted.words.xish.domish.IElement>}
        @return: A L{Mood} instance or C{None} if C{element} was not a mood
                 document or if there was no mood value element.
        """
        if element.uri != NS_MOOD or element.name != "mood":
            return None

        value = None
        text = None

        for child in element.elements():
            if child.uri != NS_MOOD:
                continue

            if child.name == "text":
                text = str(child)
            else:
                value = child.name

        if value:
            return Mood(value, text)
        else:
            return None

    fromXml = classmethod(fromXml)


class Tune:
    """
    User tune.

    This represents a user's mood, as defined in
    U{XEP-0118<http://xmpp.org/extensions/xep-0118.html>}.

    @ivar artist: The artist or performer of the song or piece.
    @type artist: L{str}
    @ivar length: The duration of the song or piece in seconds.
    @type length: L{int}
    @ivar source: The collection (e.g. album) or other source.
    @type source: L{str}
    @ivar title: The title of the song or piece
    @type title: L{str}
    @ivar track: A unique identifier for the tune; e.g. the track number within
                 the collection or the specific URI for the object.
    @type track: L{str}
    @ivar uri: A URI pointing to information about the song, collection, or
               artist.
    @type uri: L{str}

    """

    artist = None
    length = None
    source = None
    title = None
    track = None
    uri = None

    def fromXml(self, element):
        """
        Get a Tune instance from an XML representation.

        This class method parses the given XML document into a L{Tune}
        instances.

        @param element: The XML tune document.
        @type element: object providing
                       L{IElement<twisted.words.xish.domish.IElement>}
        @return: A L{Tune} instance or C{None} if C{element} was not a tune
                 document.
        """
        if element.uri != NS_TUNE or element.name != "tune":
            return None

        tune = Tune()

        for child in element.elements():
            if child.uri != NS_TUNE:
                continue

            if child.name in ("artist", "source", "title", "track", "uri"):
                setattr(tune, child.name, str(child))
            elif child.name == "length":
                try:
                    tune.length = int(str(child))
                except ValueError:
                    pass

        return tune

    fromXml = classmethod(fromXml)



# -*- coding: utf-8 -*-
# -*- test-case-name: wokkel.test.test_mam -*-
#
# Copyright (c) Adrien Cossa, 2016
# Copyright (c) Jérôme Poisson, 2018-2026
# See LICENSE for details.

"""
XMPP Message Archive Management protocol.

This protocol is specified in
U{XEP-0313<http://xmpp.org/extensions/xep-0313.html>}.
"""

from dateutil import tz

from zope.interface import implementer
from zope.interface import Interface

from twisted.words.protocols.jabber import xmlstream
from twisted.words.xish import domish
from twisted.words.protocols.jabber import jid
from twisted.words.protocols.jabber import error
from twisted.internet import defer
from twisted.python import log

from wokkel import subprotocols
from wokkel import disco
from wokkel import data_form
from wokkel import delay

from . import rsm

NS_MAM = "urn:xmpp:mam:2"
NS_FORWARD = "urn:xmpp:forward:0"
NS_ORDER_BY = "urn:xmpp:order-by:0"

FIELDS_REQUEST = "/iq[@type='get']/query[@xmlns='%s']" % NS_MAM
ARCHIVE_REQUEST = "/iq[@type='set']/query[@xmlns='%s']" % NS_MAM
PREFS_GET_REQUEST = "/iq[@type='get']/prefs[@xmlns='%s']" % NS_MAM
PREFS_SET_REQUEST = "/iq[@type='set']/prefs[@xmlns='%s']" % NS_MAM

# TODO: add the tests!


class MAMError(error.StanzaError):
    """
    MAM error.
    """

    def __init__(self, text=None):
        error.StanzaError.__init__(self, "bad-request", text=text)


class Unsupported(MAMError):
    def __init__(self, feature, text=None):
        self.feature = feature
        MAMError.__init__(self, "feature-not-implemented", "unsupported", feature, text)

    def __str__(self):
        message = MAMError.__str__(self)
        message += ", feature %r" % self.feature
        return message


class MAMRequest(object):
    """
    A Message Archive Management <query/> request.

    @ivar form: Data Form specifing the filters.
    @itype form: L{data_form.Form}

    @ivar rsm: RSM request instance.
    @itype rsm: L{rsm.RSMRequest}

    @ivar node: pubsub node id if querying a pubsub node, else None.
    @itype node: C{unicode}

    @ivar query_id: id to use to track the query
    @itype query_id: C{unicode}

    @param orderBy: Keys to order by
    @type orderBy: L{list} of L{unicode}
    """

    # FIXME: should be based on generic.Stanza

    def __init__(
        self,
        form=None,
        rsm_=None,
        node=None,
        query_id=None,
        sender=None,
        recipient=None,
        orderBy=None,
    ):
        if form is not None:
            assert form.formType == "submit"
            assert form.formNamespace == NS_MAM
        self.form = form
        self.rsm = rsm_
        self.node = node
        self.query_id = query_id
        self.sender = sender
        self.recipient = recipient
        self.orderBy = orderBy

    @classmethod
    def fromElement(cls, iq):
        """Parse the DOM representation of a MAM <query/> request.

        @param iq: <iq/> element containing a MAM <query/>.
        @type iq: L{Element<twisted.words.xish.domish.Element>}

        @return: MAMRequest instance.
        @rtype: L{MAMRequest}
        """
        sender = jid.JID(iq.getAttribute("from"))
        recipient = jid.JID(iq.getAttribute("to"))
        try:
            query = next(iq.elements(NS_MAM, "query"))
        except StopIteration:
            raise MAMError("Can't find MAM <query/> in element")
        form = data_form.findForm(query, NS_MAM)
        try:
            rsm_request = rsm.RSMRequest.fromElement(query)
        except rsm.RSMNotFoundError:
            rsm_request = None
        node = query.getAttribute("node")
        query_id = query.getAttribute("queryid")

        orderBy = []
        for element in query.elements(NS_ORDER_BY, "order"):
            orderBy.append(element["by"])

        return MAMRequest(form, rsm_request, node, query_id, sender, recipient, orderBy)

    def toElement(self):
        """
        Return the DOM representation of this RSM <query/> request.

        @rtype: L{Element<twisted.words.xish.domish.Element>}
        """
        mam_elt = domish.Element((NS_MAM, "query"))
        if self.node is not None:
            mam_elt["node"] = self.node
        if self.query_id is not None:
            mam_elt["queryid"] = self.query_id
        if self.form is not None:
            mam_elt.addChild(self.form.toElement())
        if self.rsm is not None:
            mam_elt.addChild(self.rsm.toElement())
        if self.orderBy is not None:
            for by_attr in self.orderBy:
                order_elt = mam_elt.addElement((NS_ORDER_BY, "order"))
                order_elt["by"] = by_attr

        return mam_elt

    def render(self, parent):
        """Embed the DOM representation of this MAM request in the given element.

        @param parent: parent IQ element.
        @type parent: L{Element<twisted.words.xish.domish.Element>}

        @return: MAM request element.
        @rtype: L{Element<twisted.words.xish.domish.Element>}
        """
        assert parent.name == "iq"
        mam_elt = self.toElement()
        parent.addChild(mam_elt)
        return mam_elt


class MAMPrefs(object):
    """
    A Message Archive Management <prefs/> request.

    @param default: A value in ('always', 'never', 'roster').
    @type : C{unicode} or C{None}

    @param always (list): A list of JID instances.
    @type always: C{list}

    @param never (list): A list of JID instances.
    @type never: C{list}
    """

    def __init__(self, default=None, always=None, never=None):
        if default is not None:
            # default must be defined in response, but can be empty in request (see http://xmpp.org/extensions/xep-0313.html#config)
            assert default in ("always", "never", "roster")
        self.default = default
        if always is not None:
            assert isinstance(always, list)
        else:
            always = []
        self.always = always
        if never is not None:
            assert isinstance(never, list)
        else:
            never = []
        self.never = never

    @classmethod
    def fromElement(cls, prefs_elt):
        """Parse the DOM representation of a MAM <prefs/> request.

        @param prefs_elt: MAM <prefs/> request element.
        @type prefs_elt: L{Element<twisted.words.xish.domish.Element>}

        @return: MAMPrefs instance.
        @rtype: L{MAMPrefs}
        """
        if prefs_elt.uri != NS_MAM or prefs_elt.name != "prefs":
            raise MAMError("Element provided is not a MAM <prefs/> request")
        try:
            default = prefs_elt["default"]
        except KeyError:
            # FIXME: return proper error here
            raise MAMError("Element provided is not a valid MAM <prefs/> request")

        prefs = {}
        for attr in ("always", "never"):
            prefs[attr] = []
            try:
                pref = next(prefs_elt.elements(NS_MAM, attr))
            except StopIteration:
                # FIXME: return proper error here
                raise MAMError("Element provided is not a valid MAM <prefs/> request")
            else:
                for jid_s in pref.elements(NS_MAM, "jid"):
                    prefs[attr].append(jid.JID(jid_s))
        return MAMPrefs(default, **prefs)

    def toElement(self):
        """
        Return the DOM representation of this RSM <prefs/>request.

        @rtype: L{Element<twisted.words.xish.domish.Element>}
        """
        mam_elt = domish.Element((NS_MAM, "prefs"))
        if self.default:
            mam_elt["default"] = self.default
        for attr in ("always", "never"):
            attr_elt = mam_elt.addElement(attr)
            jids = getattr(self, attr)
            for jid_ in jids:
                attr_elt.addElement("jid", content=jid_.full())
        return mam_elt

    def render(self, parent):
        """Embed the DOM representation of this MAM request in the given element.

        @param parent: parent IQ element.
        @type parent: L{Element<twisted.words.xish.domish.Element>}

        @return: MAM request element.
        @rtype: L{Element<twisted.words.xish.domish.Element>}
        """
        assert parent.name == "iq"
        mam_elt = self.toElement()
        parent.addChild(mam_elt)
        return mam_elt


class MAMClient(subprotocols.XMPPHandler):
    """
    MAM client.

    This handler implements the protocol for sending out MAM requests.
    """

    def queryArchive(self, mam_query, service=None, sender=None):
        """Query a user, MUC or pubsub archive.

        @param mam_query: query to use
        @type form: L{MAMRequest}

        @param service: Entity offering the MAM service (None for user server).
        @type service: L{JID<twisted.words.protocols.jabber.jid.JID>}

        @param sender: Optional sender address.
        @type sender: L{JID<twisted.words.protocols.jabber.jid.JID>}

        @return: A deferred that fires upon receiving a response.
        @rtype: L{Deferred<twisted.internet.defer.Deferred>}
        """
        iq = xmlstream.IQ(self.xmlstream, "set")
        mam_query.render(iq)
        if sender is not None:
            iq["from"] = str(sender)
        return iq.send(to=service.full() if service else None)

    def queryFields(self, service=None, sender=None):
        """Ask the server about supported fields.

        @param service: Entity offering the MAM service (None for user archives).
        @type service: L{JID<twisted.words.protocols.jabber.jid.JID>}

        @param sender: Optional sender address.
        @type sender: L{JID<twisted.words.protocols.jabber.jid.JID>}

        @return: data Form with the fields, or None if not found
        @rtype: L{Deferred<twisted.internet.defer.Deferred>}
        """
        # http://xmpp.org/extensions/xep-0313.html#query-form
        iq = xmlstream.IQ(self.xmlstream, "get")
        MAMRequest().render(iq)
        if sender is not None:
            iq["from"] = str(sender)
        d = iq.send(to=service.full() if service else None)
        d.addCallback(lambda iq_result: next(iq_result.elements(NS_MAM, "query")))
        d.addCallback(data_form.findForm, NS_MAM)
        return d

    def queryPrefs(self, service=None, sender=None):
        """Retrieve the current user preferences.

        @param service: Entity offering the MAM service (None for user archives).
        @type service: L{JID<twisted.words.protocols.jabber.jid.JID>}

        @param sender: Optional sender address.
        @type sender: L{JID<twisted.words.protocols.jabber.jid.JID>}

        @return: A deferred that fires upon receiving a response.
        @rtype: L{Deferred<twisted.internet.defer.Deferred>}
        """
        # http://xmpp.org/extensions/xep-0313.html#prefs
        iq = xmlstream.IQ(self.xmlstream, "get")
        MAMPrefs().render(iq)
        if sender is not None:
            iq["from"] = str(sender)
        return iq.send(to=service.full() if service else None)

    def setPrefs(
        self, service=None, default="roster", always=None, never=None, sender=None
    ):
        """Set new user preferences.

        @param service: Entity offering the MAM service (None for user archives).
        @type service: L{JID<twisted.words.protocols.jabber.jid.JID>}

        @param default: A value in ('always', 'never', 'roster').
        @type : C{unicode}

        @param always (list): A list of JID instances.
        @type always: C{list}

        @param never (list): A list of JID instances.
        @type never: C{list}

        @param sender: Optional sender address.
        @type sender: L{JID<twisted.words.protocols.jabber.jid.JID>}

        @return: A deferred that fires upon receiving a response.
        @rtype: L{Deferred<twisted.internet.defer.Deferred>}
        """
        # http://xmpp.org/extensions/xep-0313.html#prefs
        assert default is not None
        iq = xmlstream.IQ(self.xmlstream, "set")
        MAMPrefs(default, always, never).render(iq)
        if sender is not None:
            iq["from"] = str(sender)
        return iq.send(to=service.full() if service else None)


class IMAMResource(Interface):

    def onArchiveRequest(self, mam):
        """

        @param mam: The MAM <query/> request.
        @type mam: L{MAMRequest<wokkel.mam.MAMRequest>}

        @return: A tuple of (msg_data, rsm_elt, attributes).
            msg_data: list of (id, element, date) tuples
            rsm_elt: RSM response element or None
            attributes: dict of attributes for the <fin/> element
        @rtype: L{tuple}
        """

    def onPrefsGetRequest(self, requestor):
        """

        @param requestor: JID of the requestor.
        @type requestor: L{JID<twisted.words.protocols.jabber.jid.JID>}

        @return: The current settings.
        @rtype: L{wokkel.mam.MAMPrefs}
        """

    def onPrefsSetRequest(self, prefs, requestor):
        """

        @param prefs: The new settings to set.
        @type prefs: L{wokkel.mam.MAMPrefs}

        @param requestor: JID of the requestor.
        @type requestor: L{JID<twisted.words.protocols.jabber.jid.JID>}

        @return: The new current settings.
        @rtype: L{wokkel.mam.MAMPrefs}
        """


class IMAMService(Interface):
    """
    Interface for XMPP MAM service.
    """

    def addFilter(self, field):
        """
        Add a new filter for querying MAM archive.

        @param field: data form field of the filter
        @type field: L{Form<wokkel.data_form.Field>}
        """


@implementer(IMAMService, disco.IDisco)
class MAMService(subprotocols.XMPPHandler, subprotocols.IQHandlerMixin):
    """
    Protocol implementation for a MAM service.

    This handler waits for XMPP Ping requests and sends a response.
    """

    _request_class = MAMRequest

    iqHandlers = {
        FIELDS_REQUEST: "_onFieldsRequest",
        ARCHIVE_REQUEST: "_onArchiveRequest",
        PREFS_GET_REQUEST: "_onPrefsGetRequest",
        PREFS_SET_REQUEST: "_onPrefsSetRequest",
    }

    _mainFilters = {
        "start": {
            "fieldType": "text-single",
            "var": "start",
            "label": "Starting time",
            "desc": "Starting time a the result period.",
        },
        "end": {
            "fieldType": "text-single",
            "var": "end",
            "label": "Ending time",
            "desc": "Ending time of the result period.",
        },
        "with": {
            "fieldType": "jid-single",
            "var": "with",
            "label": "Entity",
            "desc": "Entity against which to match message.",
        },
    }

    def __init__(self, resource):
        """
        @param resource: instance implementing IMAMResource
        @type resource: L{object}
        """
        self.resource = resource
        self.extra_fields = {}

    def connectionInitialized(self):
        """
        Called when the XML stream has been initialized.

        This sets up an observer for incoming ping requests.
        """
        self.xmlstream.addObserver(FIELDS_REQUEST, self.handleRequest)
        self.xmlstream.addObserver(ARCHIVE_REQUEST, self.handleRequest)
        self.xmlstream.addObserver(PREFS_GET_REQUEST, self.handleRequest)
        self.xmlstream.addObserver(PREFS_SET_REQUEST, self.handleRequest)

    def addFilter(self, field):
        """
        Add a new filter for querying MAM archive.

        @param field: data form field of the filter
        @type field: L{Form<wokkel.data_form.Field>}
        """
        self.extra_fields[field.var] = field

    def _onFieldsRequest(self, iq):
        """
        Called when a fields request has been received.

        This immediately replies with a result response.
        """
        iq.handled = True
        query = domish.Element((NS_MAM, "query"))
        query.addChild(
            buildForm(extra_fields=self.extra_fields).toElement(), formType="form"
        )
        return query

    def _onArchiveRequest(self, iq):
        """
        Called when a message archive request has been received.

        This will send the list of archived message and return MAM <fin> element
        @return: The MAM <fin> element
        @rtype: L{Element<twisted.words.xish.domish.Element>}
        """
        iq.handled = True
        mam_ = self._request_class.fromElement(iq)

        # remove unsupported filters
        unsupported_fields = []
        if mam_.form:
            for key, field in mam_.form.fields.items():
                if key not in self._mainFilters and key not in self.extra_fields:
                    log.msg("Ignored unsupported MAM filter: %s" % field)
                    unsupported_fields.append(key)
        for key in unsupported_fields:
            del mam_.form.fields[key]

        def forwardMessage(id_, elt, date):
            msg = domish.Element((None, "message"))
            msg["to"] = iq["from"]
            result = msg.addElement((NS_MAM, "result"))
            if mam_.query_id is not None:
                result["queryid"] = mam_.query_id
            result["id"] = id_
            forward = result.addElement((NS_FORWARD, "forwarded"))
            forward.addChild(delay.Delay(date).toElement())
            forward.addChild(elt)
            self.xmlstream.send(msg)

        def cb(result):
            msg_data, rsm_elt, attributes = result
            for data in msg_data:
                forwardMessage(*data)

            fin_elt = domish.Element((NS_MAM, "fin"))
            for key, value in list(attributes.items()):
                fin_elt[key] = value

            if rsm_elt is not None:
                fin_elt.addChild(rsm_elt)
            return fin_elt

        d = defer.maybeDeferred(self.resource.onArchiveRequest, mam_)
        d.addCallback(cb)
        return d

    def _onPrefsGetRequest(self, iq):
        """
        Called when a prefs get request has been received.

        This immediately replies with a result response.
        """
        iq.handled = True
        requestor = jid.JID(iq["from"])

        def cb(prefs):
            return prefs.toElement()

        d = self.resource.onPrefsGetRequest(requestor).addCallback(cb)
        return d

    def _onPrefsSetRequest(self, iq):
        """
        Called when a prefs get request has been received.

        This immediately replies with a result response.
        """
        iq.handled = True

        prefs = MAMPrefs.fromElement(iq.prefs)
        requestor = jid.JID(iq["from"])

        def cb(prefs):
            return prefs.toElement()

        d = self.resource.onPrefsSetRequest(prefs, requestor).addCallback(cb)
        return d

    def getDiscoInfo(self, requestor, target, nodeIdentifier=""):
        if nodeIdentifier:
            return []
        return [disco.DiscoFeature(NS_MAM)]

    def getDiscoItems(self, requestor, target, nodeIdentifier=""):
        return []


def datetime2utc(datetime_obj):
    """Convert a datetime to a XEP-0082 compliant UTC datetime.

    @param datetime_obj: Offset-aware timestamp to convert.
    @type datetime_obj: L{datetime<datetime.datetime>}

    @return: The datetime converted to UTC.
    @rtype: C{unicode}
    """
    stampFormat = "%Y-%m-%dT%H:%M:%SZ"
    return datetime_obj.astimezone(tz.tzutc()).strftime(stampFormat)


def buildForm(start=None, end=None, with_jid=None, extra_fields=None, formType="submit"):
    """Prepare a Data Form for MAM.

    @param start: Offset-aware timestamp to filter out older messages.
    @type start: L{datetime<datetime.datetime>}

    @param end: Offset-aware timestamp to filter out later messages.
    @type end: L{datetime<datetime.datetime>}

    @param with_jid: JID against which to match messages.
    @type with_jid: L{JID<twisted.words.protocols.jabber.jid.JID>}

    @param extra_fields: list of extra data form fields that are not defined by the
        specification.
    @type: C{list}

    @param formType: The type of the Data Form ('submit' or 'form').
    @type formType: C{unicode}

    @return: XEP-0004 Data Form object.
    @rtype: L{Form<wokkel.data_form.Form>}
    """
    form = data_form.Form(formType, formNamespace=NS_MAM)

    if formType == "form":
        for kwargs in MAMService._mainFilters.values():
            form.addField(data_form.Field(**kwargs))
    elif formType == "submit":
        if start:
            form.addField(data_form.Field(var="start", value=datetime2utc(start)))
        if end:
            form.addField(data_form.Field(var="end", value=datetime2utc(end)))
        if with_jid:
            form.addField(
                data_form.Field(fieldType="jid-single", var="with", value=with_jid.full())
            )

    if extra_fields is not None:
        for field in extra_fields:
            form.addField(field)

    return form



# Copyright (c) Ralph Meijer.
# See LICENSE for details.

"""
Unit test helpers.
"""

from __future__ import division, absolute_import

from twisted.internet import defer
from twisted.words.xish import xpath
from twisted.words.xish.utility import EventDispatcher

from wokkel.generic import parseXml
from wokkel.subprotocols import StreamManager


class XmlStreamStub(object):
    """
    Stub for testing objects that communicate through XML Streams.

    Instances of this stub hold an object in L{xmlstream} that acts like an
    L{XmlStream<twisted.words.xish.xmlstream.XmlStream} after connection stream
    initialization. Stanzas can be sent through it by calling its C{send}
    method with an object implementing
    L{IElement<twisted.words.xish.domish.IElement>} as its first argument.
    These appear in sequence in the L{output} instance variable of the stub.

    For the reverse direction, stanzas passed to L{send} of the stub, will be
    dispatched in the stubbed XmlStream as if it was received over the wire, so
    that registered observers will be called.

    Example::

        >>> stub = XmlStreamStub()
        >>> stub.xmlstream.send(domish.Element((None, 'presence')))
        >>> stub.output[-1].toXml()
        u'<presence/>'
        >>> def cb(stanza):
        ...     print("Got: %r" stanza.toXml())
        >>> stub.xmlstream.addObserver('/presence')
        >>> stub.send(domish.Element((None, 'presence')))
        Got: u'<presence/>'

    @ivar xmlstream: Stubbed XML Stream.
    @type xmlstream: L{EventDispatcher}
    @ivar output: List of stanzas sent to the XML Stream.
    @type output: L{list}
    """

    def __init__(self):
        self.output = []
        self.xmlstream = EventDispatcher()
        self.xmlstream.send = self.output.append

    def send(self, obj):
        """
        Pass an element to the XML Stream as if received.

        @param obj: Element to be dispatched to C{self.xmlstream}.
        @type obj: object implementing
                   L{IElement<twisted.words.xish.domish.IElement>}.
        """
        self.xmlstream.dispatch(obj)


class TestableRequestHandlerMixin(object):
    """
    Mixin for testing XMPPHandlers that process iq requests.

    Handlers that use L{wokkel.subprotocols.IQHandlerMixin} define a
    C{iqHandlers} attribute that lists the handlers to be called for iq
    requests. This mixin provides L{handleRequest} to mimic the handler
    processing for easier testing.
    """

    def handleRequest(self, xml):
        """
        Find a handler and call it directly.

        @param xml: XML stanza that may yield a handler being called.
        @type xml: L{str}.
        @return: Deferred that fires with the result of a handler for this
                 stanza. If no handler was found, the deferred has its errback
                 called with a C{NotImplementedError} exception.
        """
        handler = None
        iq = parseXml(xml)
        for queryString, method in self.service.iqHandlers.items():
            if xpath.internQuery(queryString).matches(iq):
                handler = getattr(self.service, method)

        if handler:
            d = defer.maybeDeferred(handler, iq)
        else:
            d = defer.fail(NotImplementedError())

        return d


class TestableStreamManager(StreamManager):
    """
    Stream manager for testing subprotocol handlers.
    """

    def __init__(self, reactor=None):
        class DummyFactory(object):
            def addBootstrap(self, event, fn):
                pass

        factory = DummyFactory()
        StreamManager.__init__(self, factory, reactor)
        self.stub = XmlStreamStub()
        self._connected(self.stub.xmlstream)
        self._authd(self.stub.xmlstream)



# Copyright (c) Ralph Meijer.
# See LICENSE for details.

"""
Tests for L{wokkel.format}
"""

from __future__ import division, absolute_import

from twisted.trial import unittest
from twisted.words.xish import domish

from wokkel import formats


class MoodTests(unittest.TestCase):
    """
    Tests for L{formats.Mood}.
    """

    def test_fromXml(self):
        """
        Moods are parsed from Elements.
        """
        element = domish.Element((formats.NS_MOOD, "mood"))
        element.addElement("happy")
        element.addElement("text", content="Really happy!")
        mood = formats.Mood.fromXml(element)
        self.assertEquals("happy", mood.value)
        self.assertEquals("Really happy!", mood.text)


class TuneTests(unittest.TestCase):
    """
    Tests for L{formats.Tune}.
    """

    def test_fromXmlTrack(self):
        """
        The track filed in a user tune is parsed.
        """
        element = domish.Element((formats.NS_TUNE, "tune"))
        element.addElement("track", content="The Power")
        tune = formats.Tune.fromXml(element)
        self.assertEquals("The Power", tune.track)

    def test_fromXmlLength(self):
        """
        The length filed in a user tune is parsed as an int.
        """
        element = domish.Element((formats.NS_TUNE, "tune"))
        element.addElement("length", content="322")
        tune = formats.Tune.fromXml(element)
        self.assertEquals(322, tune.length)
